"""
Easy fillable form

Contrib - Tim Ashley Jenkins 2018

This module contains a function that calls an easily customizable EvMenu - this
menu presents the player with a fillable form, with fields that can be filled
out in any order. Each field's value can be verified, with the function
allowing easy checks for text and integer input, minimum and maximum values /
character lengths, or can even be verified by a custom function. Once the form
is submitted, the form's data is submitted as a dictionary to any callable of
your choice.

The function that initializes the fillable form menu is fairly simple, and
includes the caller, the template for the form, and the callback(caller, result)
to which the form data will be sent to upon submission.

    init_fill_field(formtemplate, caller, formcallback)

Form templates are defined as a list of dictionaries - each dictionary
represents a field in the form, and contains the data for the field's name and
behavior. For example, this basic form template will allow a player to fill out
a brief character profile:

    PROFILE_TEMPLATE = [
    {"fieldname":"Name", "fieldtype":"text"},
    {"fieldname":"Age", "fieldtype":"number"},
    {"fieldname":"History", "fieldtype":"text"},
    ]

This will present the player with an EvMenu showing this basic form:

      Name:
       Age:
   History:

While in this menu, the player can assign a new value to any field with the
syntax <field> = <new value>, like so:

    > name = Ashley
    Field 'Name' set to: Ashley

Typing 'look' by itself will show the form and its current values.

    > look

      Name: Ashley
       Age:
   History:

Number fields require an integer input, and will reject any text that can't
be converted into an integer.

    > age = youthful
    Field 'Age' requires a number.
    > age = 31
    Field 'Age' set to: 31

Form data is presented as an EvTable, so text of any length will wrap cleanly.

    > history = EVERY MORNING I WAKE UP AND OPEN PALM SLAM[...]
    Field 'History' set to: EVERY MORNING I WAKE UP AND[...]
    > look

      Name: Ashley
       Age: 31
   History: EVERY MORNING I WAKE UP AND OPEN PALM SLAM A VHS INTO THE SLOT.
            IT'S CHRONICLES OF RIDDICK AND RIGHT THEN AND THERE I START DOING
            THE MOVES ALONGSIDE WITH THE MAIN CHARACTER, RIDDICK. I DO EVERY
            MOVE AND I DO EVERY MOVE HARD.

When the player types 'submit' (or your specified submit command), the menu
quits and the form's data is passed to your specified function as a dictionary,
like so:

    formdata = {"Name":"Ashley", "Age":31, "History":"EVERY MORNING I[...]"}

You can do whatever you like with this data in your function - forms can be used
to set data on a character, to help builders create objects, or for players to
craft items or perform other complicated actions with many variables involved.

The data that your form will accept can also be specified in your form template -
let's say, for example, that you won't accept ages under 18 or over 100. You can
do this by specifying "min" and "max" values in your field's dictionary:

    PROFILE_TEMPLATE = [
    {"fieldname":"Name", "fieldtype":"text"},
    {"fieldname":"Age", "fieldtype":"number", "min":18, "max":100},
    {"fieldname":"History", "fieldtype":"text"}
    ]

Now if the player tries to enter a value out of range, the form will not acept the
given value.

    > age = 10
    Field 'Age' reqiures a minimum value of 18.
    > age = 900
    Field 'Age' has a maximum value of 100.

Setting 'min' and 'max' for a text field will instead act as a minimum or
maximum character length for the player's input.

There are lots of ways to present the form to the player - fields can have default
values or show a custom message in place of a blank value, and player input can be
verified by a custom function, allowing for a great deal of flexibility. There
is also an option for 'bool' fields, which accept only a True / False input and
can be customized to represent the choice to the player however you like (E.G.
Yes/No, On/Off, Enabled/Disabled, etc.)

This module contains a simple example form that demonstrates all of the included
functionality - a command that allows a player to compose a message to another
online character and have it send after a custom delay. You can test it by
importing this module in your game's default_cmdsets.py module and adding
CmdTestMenu to your default character's command set.

FIELD TEMPLATE KEYS:
Required:
    fieldname (str): Name of the field, as presented to the player.
    fieldtype (str): Type of value required: 'text', 'number', or 'bool'.

Optional:
    max (int): Maximum character length (if text) or value (if number).
    min (int): Minimum charater length (if text) or value (if number).
    truestr (str): String for a 'True' value in a bool field.
        (E.G. 'On', 'Enabled', 'Yes')
    falsestr (str): String for a 'False' value in a bool field.
        (E.G. 'Off', 'Disabled', 'No')
    default (str): Initial value (blank if not given).
    blankmsg (str): Message to show in place of value when field is blank.
    cantclear (bool): Field can't be cleared if True.
    required (bool): If True, form cannot be submitted while field is blank.
    verifyfunc (callable): Name of a callable used to verify input - takes
        (caller, value) as arguments. If the function returns True,
        the player's input is considered valid - if it returns False,
        the input is rejected. Any other value returned will act as
        the field's new value, replacing the player's input. This
        allows for values that aren't strings or integers (such as
        object dbrefs). For boolean fields, return '0' or '1' to set
        the field to False or True.
"""
import evennia
from evennia import Command
from evennia.utils import delay, evmenu, evtable, list_to_string, logger


class FieldEvMenu(evmenu.EvMenu):
    """
    Custom EvMenu type with its own node formatter - removes extraneous lines
    """

    def node_formatter(self, nodetext, optionstext):
        """
        Formats the entirety of the node.

        Args:
            nodetext (str): The node text as returned by `self.nodetext_formatter`.
            optionstext (str): The options display as returned by `self.options_formatter`.
            caller (Object, Account or None, optional): The caller of the node.

        Returns:
            node (str): The formatted node to display.

        """
        # Only return node text, no options or separators
        return nodetext


def init_fill_field(
    formtemplate,
    caller,
    formcallback,
    pretext="",
    posttext="",
    submitcmd="submit",
    borderstyle="cells",
    formhelptext=None,
    persistent=False,
    initial_formdata=None,
):
    """
    Initializes a menu presenting a player with a fillable form - once the form
    is submitted, the data will be passed as a dictionary to your chosen
    function.

    Args:
        formtemplate (list of dicts): The template for the form's fields.
        caller (obj): Player who will be filling out the form.
        formcallback (callable): Function to pass the completed form's data to.

    Options:
        pretext (str): Text to put before the form in the menu.
        posttext (str): Text to put after the form in the menu.
        submitcmd (str): Command used to submit the form.
        borderstyle (str): Form's EvTable border style.
        formhelptext (str): Help text for the form menu (or default is provided).
        persistent (bool): Whether to make the EvMenu persistent across reboots.
        initial_formdata (dict): Initial data for the form - a blank form with
            defaults specified in the template will be generated otherwise.
            In the case of a form used to edit properties on an object or a
            similar application, you may want to generate the initial form
            data dynamically before calling init_fill_field.
    """

    # Initialize form data from the template if none provided
    formdata = form_template_to_dict(formtemplate)
    if initial_formdata:
        formdata = initial_formdata

    # Provide default help text if none given
    if formhelptext is None:
        formhelptext = (
            "Available commands:|/"
            "|w<field> = <new value>:|n Set given field to new value, replacing the old value|/"
            "|wclear <field>:|n Clear the value in the given field, making it blank|/"
            "|wlook|n: Show the form's current values|/"
            "|whelp|n: Display this help screen|/"
            "|wquit|n: Quit the form menu without submitting|/"
            "|w%s|n: Submit this form and quit the menu" % submitcmd
        )

    # Pass kwargs to store data needed in the menu
    kwargs = {
        "formdata": formdata,
        "formtemplate": formtemplate,
        "formcallback": formcallback,
        "pretext": pretext,
        "posttext": posttext,
        "submitcmd": submitcmd,
        "borderstyle": borderstyle,
        "formhelptext": formhelptext,
    }

    # Initialize menu of selections
    FieldEvMenu(
        caller,
        "evennia.contrib.utils.fieldfill.fieldfill",
        startnode="menunode_fieldfill",
        auto_look=False,
        persistent=persistent,
        **kwargs,
    )


def menunode_fieldfill(caller, raw_string, **kwargs):
    """
    This is an EvMenu node, which calls itself over and over in order to
    allow a player to enter values into a fillable form. When the form is
    submitted, the form data is passed to a callback as a dictionary.
    """

    # Retrieve menu info - taken from ndb if not persistent or db if persistent
    if not caller.db._menutree:
        formdata = caller.ndb._menutree.formdata
        formtemplate = caller.ndb._menutree.formtemplate
        formcallback = caller.ndb._menutree.formcallback
        pretext = caller.ndb._menutree.pretext
        posttext = caller.ndb._menutree.posttext
        submitcmd = caller.ndb._menutree.submitcmd
        borderstyle = caller.ndb._menutree.borderstyle
        formhelptext = caller.ndb._menutree.formhelptext
    else:
        formdata = caller.db._menutree.formdata
        formtemplate = caller.db._menutree.formtemplate
        formcallback = caller.db._menutree.formcallback
        pretext = caller.db._menutree.pretext
        posttext = caller.db._menutree.posttext
        submitcmd = caller.db._menutree.submitcmd
        borderstyle = caller.db._menutree.borderstyle
        formhelptext = caller.db._menutree.formhelptext

    # Syntax error
    syntax_err = (
        "Syntax: <field> = <new value>|/Or: clear <field>, help, look, quit|/'%s' to submit form"
        % submitcmd
    )

    # Display current form data
    text = (
        display_formdata(
            formtemplate, formdata, pretext=pretext, posttext=posttext, borderstyle=borderstyle
        ),
        formhelptext,
    )
    options = {"key": "_default", "goto": "menunode_fieldfill"}

    if raw_string:
        # Test for given 'submit' command
        if raw_string.lower().strip() == submitcmd:
            # Test to see if any blank fields are required
            blank_and_required = []
            for field in formtemplate:
                if "required" in field.keys():
                    # If field is required but current form data for field is blank
                    if field["required"] is True and formdata[field["fieldname"]] is None:
                        # Add to blank and required fields
                        blank_and_required.append(field["fieldname"])
            if len(blank_and_required) > 0:
                # List the required fields left empty to the player
                caller.msg(
                    "The following blank fields require a value: %s"
                    % list_to_string(blank_and_required)
                )
                text = (None, formhelptext)
                return text, options

            # If everything checks out, pass form data to the callback and end the menu!
            try:
                formcallback(caller, formdata)
            except Exception:
                logger.log_trace("Error in fillable form callback.")
            return None, None

        # Test for 'look' command
        if raw_string.lower().strip() == "look" or raw_string.lower().strip() == "l":
            return text, options

        # Test for 'clear' command
        cleartest = raw_string.lower().strip().split(" ", 1)
        if cleartest[0].lower() == "clear":
            text = (None, formhelptext)
            if len(cleartest) < 2:
                caller.msg(syntax_err)
                return text, options
            matched_field = None

            for key in formdata.keys():
                if cleartest[1].lower() in key.lower():
                    matched_field = key

            if not matched_field:
                caller.msg("Field '%s' does not exist!" % cleartest[1])
                text = (None, formhelptext)
                return text, options

            # Test to see if field can be cleared
            for field in formtemplate:
                if field["fieldname"] == matched_field and "cantclear" in field.keys():
                    if field["cantclear"] is True:
                        caller.msg("Field '%s' can't be cleared!" % matched_field)
                        text = (None, formhelptext)
                        return text, options

            # Clear the field
            formdata.update({matched_field: None})
            caller.ndb._menutree.formdata = formdata
            caller.msg("Field '%s' cleared." % matched_field)
            return text, options

        if "=" not in raw_string:
            text = (None, formhelptext)
            caller.msg(syntax_err)
            return text, options

        # Extract field name and new field value
        entry = raw_string.split("=", 1)
        fieldname = entry[0].strip()
        newvalue = entry[1].strip()

        # Syntax error if field name is too short or blank
        if len(fieldname) < 1:
            caller.msg(syntax_err)
            text = (None, formhelptext)
            return text, options

        # Attempt to match field name to field in form data
        matched_field = None
        for key in formdata.keys():
            if fieldname.lower() in key.lower():
                matched_field = key

        # No matched field
        if matched_field is None:
            caller.msg("Field '%s' does not exist!" % fieldname)
            text = (None, formhelptext)
            return text, options

        # Set new field value if match
        # Get data from template
        fieldtype = None
        max_value = None
        min_value = None
        truestr = "True"
        falsestr = "False"
        verifyfunc = None
        for field in formtemplate:
            if field["fieldname"] == matched_field:
                fieldtype = field["fieldtype"]
                if "max" in field.keys():
                    max_value = field["max"]
                if "min" in field.keys():
                    min_value = field["min"]
                if "truestr" in field.keys():
                    truestr = field["truestr"]
                if "falsestr" in field.keys():
                    falsestr = field["falsestr"]
                if "verifyfunc" in field.keys():
                    verifyfunc = field["verifyfunc"]

        # Field type text verification
        if fieldtype == "text":
            # Test for max/min
            if max_value is not None:
                if len(newvalue) > max_value:
                    caller.msg(
                        "Field '%s' has a maximum length of %i characters."
                        % (matched_field, max_value)
                    )
                    text = (None, formhelptext)
                    return text, options
            if min_value is not None:
                if len(newvalue) < min_value:
                    caller.msg(
                        "Field '%s' reqiures a minimum length of %i characters."
                        % (matched_field, min_value)
                    )
                    text = (None, formhelptext)
                    return text, options

        # Field type number verification
        if fieldtype == "number":
            try:
                newvalue = int(newvalue)
            except:
                caller.msg("Field '%s' requires a number." % matched_field)
                text = (None, formhelptext)
                return text, options
            # Test for max/min
            if max_value is not None:
                if newvalue > max_value:
                    caller.msg("Field '%s' has a maximum value of %i." % (matched_field, max_value))
                    text = (None, formhelptext)
                    return text, options
            if min_value is not None:
                if newvalue < min_value:
                    caller.msg(
                        "Field '%s' reqiures a minimum value of %i." % (matched_field, min_value)
                    )
                    text = (None, formhelptext)
                    return text, options

        # Field type bool verification
        if fieldtype == "bool":
            if newvalue.lower() != truestr.lower() and newvalue.lower() != falsestr.lower():
                caller.msg(
                    "Please enter '%s' or '%s' for field '%s'." % (truestr, falsestr, matched_field)
                )
                text = (None, formhelptext)
                return text, options
            if newvalue.lower() == truestr.lower():
                newvalue = True
            elif newvalue.lower() == falsestr.lower():
                newvalue = False

        # Call verify function if present
        if verifyfunc:
            if verifyfunc(caller, newvalue) is False:
                # No error message is given - should be provided by verifyfunc
                text = (None, formhelptext)
                return text, options
            elif verifyfunc(caller, newvalue) is not True:
                newvalue = verifyfunc(caller, newvalue)
                # Set '0' or '1' to True or False if the field type is bool
                if fieldtype == "bool":
                    if newvalue == 0:
                        newvalue = False
                    elif newvalue == 1:
                        newvalue = True

        # If everything checks out, update form!!
        formdata.update({matched_field: newvalue})
        caller.ndb._menutree.formdata = formdata

        # Account for truestr and falsestr when updating a boolean form
        announced_newvalue = newvalue
        if newvalue is True:
            announced_newvalue = truestr
        elif newvalue is False:
            announced_newvalue = falsestr

        # Announce the new value to the player
        caller.msg("Field '%s' set to: %s" % (matched_field, str(announced_newvalue)))
        text = (None, formhelptext)

    return text, options


def form_template_to_dict(formtemplate):
    """
    Initializes a dictionary of form data from the given list-of-dictionaries
    form template, as formatted above.

    Args:
        formtemplate (list of dicts): Tempate for the form to be initialized.

    Returns:
        formdata (dict): Dictionary of initalized form data.
    """
    formdata = {}

    for field in formtemplate:
        # Value is blank by default
        fieldvalue = None
        if "default" in field:
            # Add in default value if present
            fieldvalue = field["default"]
        formdata.update({field["fieldname"]: fieldvalue})

    return formdata


def display_formdata(formtemplate, formdata, pretext="", posttext="", borderstyle="cells"):
    """
    Displays a form's current data as a table. Used in the form menu.

    Args:
        formtemplate (list of dicts): Template for the form
        formdata (dict): Form's current data

    Options:
        pretext (str): Text to put before the form table.
        posttext (str): Text to put after the form table.
        borderstyle (str): EvTable's border style.
    """

    formtable = evtable.EvTable(border=borderstyle, valign="t", maxwidth=80)
    field_name_width = 5

    for field in formtemplate:
        new_fieldname = None
        new_fieldvalue = None
        # Get field name
        new_fieldname = "|w" + field["fieldname"] + ":|n"
        if len(field["fieldname"]) + 5 > field_name_width:
            field_name_width = len(field["fieldname"]) + 5
        # Get field value
        if formdata[field["fieldname"]] is not None:
            new_fieldvalue = str(formdata[field["fieldname"]])
        # Use blank message if field is blank and once is present
        if new_fieldvalue is None and "blankmsg" in field:
            new_fieldvalue = "|x" + str(field["blankmsg"]) + "|n"
        elif new_fieldvalue is None:
            new_fieldvalue = " "
        # Replace True and False values with truestr and falsestr from template
        if formdata[field["fieldname"]] is True and "truestr" in field:
            new_fieldvalue = field["truestr"]
        elif formdata[field["fieldname"]] is False and "falsestr" in field:
            new_fieldvalue = field["falsestr"]
        # Add name and value to table
        formtable.add_row(new_fieldname, new_fieldvalue)

    formtable.reformat_column(0, align="r", width=field_name_width)

    return pretext + "|/" + str(formtable) + "|/" + posttext


# EXAMPLE FUNCTIONS / COMMAND STARTS HERE


def verify_online_player(caller, value):
    """
    Example 'verify function' that matches player input to an online character
    or else rejects their input as invalid.

    Args:
        caller (obj): Player entering the form data.
        value (str): String player entered into the form, to be verified.

    Returns:
        matched_character (obj or False): dbref to a currently logged in
            character object - reference to the object will be stored in
            the form instead of a string. Returns False if no match is
            made.
    """
    # Get a list of sessions
    session_list = evennia.SESSION_HANDLER.get_sessions()
    char_list = []
    matched_character = None

    # Get a list of online characters
    for session in session_list:
        if not session.logged_in:
            # Skip over logged out characters
            continue
        # Append to our list of online characters otherwise
        char_list.append(session.get_puppet())

    # Match player input to a character name
    for character in char_list:
        if value.lower() == character.key.lower():
            matched_character = character

    # If input didn't match to a character
    if not matched_character:
        # Send the player an error message unique to this function
        caller.msg("No character matching '%s' is online." % value)
        # Returning False indicates the new value is not valid
        return False

    # Returning anything besides True or False will replace the player's input with the returned
    # value. In this case, the value becomes a reference to the character object. You can store data
    # besides strings and integers in the 'formdata' dictionary this way!
    return matched_character


# Form template for the example 'delayed message' form
SAMPLE_FORM = [
    {
        "fieldname": "Character",
        "fieldtype": "text",
        "max": 30,
        "blankmsg": "(Name of an online player)",
        "required": True,
        "verifyfunc": verify_online_player,
    },
    {
        "fieldname": "Delay",
        "fieldtype": "number",
        "min": 3,
        "max": 30,
        "default": 10,
        "cantclear": True,
    },
    {
        "fieldname": "Message",
        "fieldtype": "text",
        "min": 3,
        "max": 200,
        "blankmsg": "(Message up to 200 characters)",
    },
    {
        "fieldname": "Anonymous",
        "fieldtype": "bool",
        "truestr": "Yes",
        "falsestr": "No",
        "default": False,
    },
]


class CmdTestMenu(Command):
    """
    This test command will initialize a menu that presents you with a form.
    You can fill out the fields of this form in any order, and then type in
    'send' to send a message to another online player, which will reach them
    after a delay you specify.

    Usage:
       <field> = <new value>
       clear <field>
       help
       look
       quit
       send
    """

    key = "testmenu"

    def func(self):
        """
        This performs the actual command.
        """
        pretext = (
            "|cSend a delayed message to another player ---------------------------------------|n"
        )
        posttext = (
            "|c--------------------------------------------------------------------------------|n|/"
            "Syntax: type |c<field> = <new value>|n to change the values of the form. Given|/"
            "player must be currently logged in, delay is given in seconds. When you are|/"
            "finished, type '|csend|n' to send the message.|/"
        )

        init_fill_field(
            SAMPLE_FORM,
            self.caller,
            init_delayed_message,
            pretext=pretext,
            posttext=posttext,
            submitcmd="send",
            borderstyle="none",
        )


def sendmessage(obj, text):
    """
    Callback to send a message to a player.

    Args:
        obj (obj): Player to message.
        text (str): Message.
    """
    obj.msg(text)


def init_delayed_message(caller, formdata):
    """
    Initializes a delayed message, using data from the example form.

    Args:
        caller (obj): Character submitting the message.
        formdata (dict): Data from submitted form.
    """
    # Retrieve data from the filled out form.
    # We stored the character to message as an object ref using a verifyfunc
    # So we don't have to do any more searching or matching here!
    player_to_message = formdata["Character"]
    message_delay = formdata["Delay"]
    sender = str(caller)
    if formdata["Anonymous"] is True:
        sender = "anonymous"
    message = ("Message from %s: " % sender) + str(formdata["Message"])

    caller.msg("Message sent to %s!" % player_to_message)
    # Make a deferred call to 'sendmessage' above.
    delay(message_delay, sendmessage, player_to_message, message)
    return



"""
Test chargen.

"""

from unittest.mock import MagicMock, patch

from parameterized import parameterized

from evennia import create_object
from evennia.utils.test_resources import BaseEvenniaTest

from .. import chargen, enums, objects


class EvAdventureCharacterGenerationTest(BaseEvenniaTest):
    """
    Test the Character generator in the rule engine.

    """

    @patch("evennia.contrib.tutorials.evadventure.rules.randint")
    def setUp(self, mock_randint):
        super().setUp()
        mock_randint.return_value = 10
        self.chargen = chargen.TemporaryCharacterSheet()

    def test_base_chargen(self):
        self.assertEqual(self.chargen.strength, 10)  # not realistic, due to mock
        self.assertEqual(self.chargen.armor, "gambeson")
        self.assertEqual(self.chargen.shield, "shield")
        self.assertEqual(
            self.chargen.backpack, ["ration", "ration", "waterskin", "waterskin", "drill", "twine"]
        )

    def test_build_desc(self):
        self.assertEqual(
            self.chargen.desc,
            "You are scrawny with a broken face, pockmarked skin, greased hair, hoarse speech, and "
            "stained clothing. You were a Herbalist, but you were exiled and ended up a knave. You "
            "are honest but also irascible. You tend towards neutrality.",
        )

    @patch("evennia.contrib.tutorials.evadventure.chargen.spawn")
    def test_apply(self, mock_spawn):
        gambeson = create_object(objects.EvAdventureArmor, key="gambeson")
        mock_spawn.return_value = [gambeson]
        account = MagicMock()
        account.id = 2222

        character = self.chargen.apply(account)

        self.assertIn("Herbalist", character.db.desc)
        self.assertEqual(
            character.equipment.all(),
            [
                (None, enums.WieldLocation.WEAPON_HAND),
                (None, enums.WieldLocation.SHIELD_HAND),
                (None, enums.WieldLocation.TWO_HANDS),
                (gambeson, enums.WieldLocation.BODY),
                (None, enums.WieldLocation.HEAD),
            ],
        )

        gambeson.delete()
        character.delete()



# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "typeclasses",
            "0003_defaultcharacter_defaultexit_defaultguest_defaultobject_defaultplayer_defaultroom_defaultscript_dono",
        )
    ]

    operations = [
        migrations.AlterField(
            model_name="attribute",
            name="db_model",
            field=models.CharField(
                max_length=32,
                blank=True,
                help_text="Which model of object this attribute is attached to (A natural key like 'objects.objectdb'). You should not change this value unless you know what you are doing.",
                null=True,
                verbose_name="model",
                db_index=True,
            ),
        )
    ]



"""
Webclient based on websockets.

This implements a webclient with WebSockets (http://en.wikipedia.org/wiki/WebSocket)
by use of the autobahn-python package's implementation (https://github.com/crossbario/autobahn-python).
It is used together with evennia/web/media/javascript/evennia_websocket_webclient.js.

All data coming into the webclient is in the form of valid JSON on the form

`["inputfunc_name", [args], {kwarg}]`

which represents an "inputfunc" to be called on the Evennia side with *args, **kwargs.
The most common inputfunc is "text", which takes just the text input
from the command line and interprets it as an Evennia Command: `["text", ["look"], {}]`

"""
import html
import json
import re

from autobahn.exception import Disconnected
from autobahn.twisted.websocket import WebSocketServerProtocol
from django.conf import settings

from evennia.utils.ansi import parse_ansi
from evennia.utils.text2html import parse_html
from evennia.utils.utils import class_from_module, mod_import

_RE_SCREENREADER_REGEX = re.compile(
    r"%s" % settings.SCREENREADER_REGEX_STRIP, re.DOTALL + re.MULTILINE
)
_CLIENT_SESSIONS = mod_import(settings.SESSION_ENGINE).SessionStore
_UPSTREAM_IPS = settings.UPSTREAM_IPS

# Status Code 1000: Normal Closure
#   called when the connection was closed through JavaScript
CLOSE_NORMAL = WebSocketServerProtocol.CLOSE_STATUS_CODE_NORMAL

# Status Code 1001: Going Away
#   called when the browser is navigating away from the page
GOING_AWAY = WebSocketServerProtocol.CLOSE_STATUS_CODE_GOING_AWAY

_BASE_SESSION_CLASS = class_from_module(settings.BASE_SESSION_CLASS)


class WebSocketClient(WebSocketServerProtocol, _BASE_SESSION_CLASS):
    """
    Implements the server-side of the Websocket connection.

    """

    # nonce value, used to prevent the webclient from erasing the
    # webclient_authenticated_uid value of csession on disconnect
    nonce = 0

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.protocol_key = "webclient/websocket"
        self.browserstr = ""

    def get_client_session(self):
        """
        Get the Client browser session (used for auto-login based on browser session)

        Returns:
            csession (ClientSession): This is a django-specific internal representation
                of the browser session.

        """
        try:
            # client will connect with wsurl?csessid&page_id&browserid
            webarg = self.http_request_uri.split("?", 1)[1]
        except IndexError:
            # this may happen for custom webclients not caring for the
            # browser session.
            self.csessid = None
            return None
        except AttributeError:
            from evennia.utils import logger

            self.csessid = None
            logger.log_trace(str(self))
            return None

        self.csessid, *cargs = webarg.split("&", 2)
        if len(cargs) == 1:
            self.browserstr = str(cargs[0])
        elif len(cargs) == 2:
            self.page_id = str(cargs[0])
            self.browserstr = str(cargs[1])

        if self.csessid:
            return _CLIENT_SESSIONS(session_key=self.csessid)

    def onOpen(self):
        """
        This is called when the WebSocket connection is fully established.

        """
        client_address = self.transport.client
        client_address = client_address[0] if client_address else None

        if client_address in _UPSTREAM_IPS and "x-forwarded-for" in self.http_headers:
            addresses = [x.strip() for x in self.http_headers["x-forwarded-for"].split(",")]
            addresses.reverse()

            for addr in addresses:
                if addr not in _UPSTREAM_IPS:
                    client_address = addr
                    break

        self.init_session("websocket", client_address, self.factory.sessionhandler)

        csession = self.get_client_session()  # this sets self.csessid
        csessid = self.csessid
        uid = csession and csession.get("webclient_authenticated_uid", None)
        nonce = csession and csession.get("webclient_authenticated_nonce", 0)
        if uid:
            # the client session is already logged in.
            self.uid = uid
            self.nonce = nonce
            self.logged_in = True

            for old_session in self.sessionhandler.sessions_from_csessid(csessid):
                if (
                    hasattr(old_session, "websocket_close_code")
                    and old_session.websocket_close_code != CLOSE_NORMAL
                ):
                    # if we have old sessions with the same csession, they are remnants
                    self.sessid = old_session.sessid
                    self.sessionhandler.disconnect(old_session)

        browserstr = f":{self.browserstr}" if self.browserstr else ""
        self.protocol_flags["CLIENTNAME"] = f"Evennia Webclient (websocket{browserstr})"
        self.protocol_flags["UTF-8"] = True
        self.protocol_flags["OOB"] = True
        self.protocol_flags["TRUECOLOR"] = True
        self.protocol_flags["XTERM256"] = True
        self.protocol_flags["ANSI"] = True

        # watch for dead links
        self.transport.setTcpKeepAlive(1)
        # actually do the connection
        self.sessionhandler.connect(self)

    def disconnect(self, reason=None):
        """
        Generic hook for the engine to call in order to
        disconnect this protocol.

        Args:
            reason (str or None): Motivation for the disconnection.

        """
        csession = self.get_client_session()

        if csession:
            # if the nonce is different, webclient_authenticated_uid has been
            # set *before* this disconnect (disconnect called after a new client
            # connects, which occurs in some 'fast' browsers like Google Chrome
            # and Mobile Safari)
            if csession.get("webclient_authenticated_nonce", 0) == self.nonce:
                csession["webclient_authenticated_uid"] = None
                csession["webclient_authenticated_nonce"] = 0
                csession.save()
            self.logged_in = False

        self.sessionhandler.disconnect(self)
        # autobahn-python:
        # 1000 for a normal close, 1001 if the browser window is closed,
        # 3000-4999 for app. specific,
        # in case anyone wants to expose this functionality later.
        #
        # sendClose() under autobahn/websocket/interfaces.py
        self.sendClose(CLOSE_NORMAL, reason)

    def onClose(self, wasClean, code=None, reason=None):
        """
        This is executed when the connection is lost for whatever
        reason. it can also be called directly, from the disconnect
        method.

        Args:
            wasClean (bool): ``True`` if the WebSocket was closed cleanly.
            code (int or None): Close status as sent by the WebSocket peer.
            reason (str or None): Close reason as sent by the WebSocket peer.

        """
        if code == CLOSE_NORMAL or code == GOING_AWAY:
            self.disconnect(reason)
        else:
            self.websocket_close_code = code

    def onMessage(self, payload, isBinary):
        """
        Callback fired when a complete WebSocket message was received.

        Args:
            payload (bytes): The WebSocket message received.
            isBinary (bool): Flag indicating whether payload is binary or
                             UTF-8 encoded text.

        """
        cmdarray = json.loads(str(payload, "utf-8"))
        if cmdarray:
            self.data_in(**{cmdarray[0]: [cmdarray[1], cmdarray[2]]})

    def sendLine(self, line):
        """
        Send data to client.

        Args:
            line (str): Text to send.

        """
        try:
            return self.sendMessage(line.encode())
        except Disconnected:
            # this can happen on an unclean close of certain browsers.
            # it means this link is actually already closed.
            self.disconnect(reason="Browser already closed.")

    def at_login(self):
        csession = self.get_client_session()
        if csession:
            csession["webclient_authenticated_uid"] = self.uid
            csession.save()

    def data_in(self, **kwargs):
        """
        Data User > Evennia.

        Args:
            text (str): Incoming text.
            kwargs (any): Options from protocol.

        Notes:
            At initilization, the client will send the special
            'csessid' command to identify its browser session hash
            with the Evennia side.

            The websocket client will also pass 'websocket_close' command
            to report that the client has been closed and that the
            session should be disconnected.

            Both those commands are parsed and extracted already at
            this point.

        """
        if "websocket_close" in kwargs:
            self.disconnect()
            return

        self.sessionhandler.data_in(self, **kwargs)

    def send_text(self, *args, **kwargs):
        """
        Send text data. This will pre-process the text for
        color-replacement, conversion to html etc.

        Args:
            text (str): Text to send.

        Keyword Args:
            options (dict): Options-dict with the following keys understood:
                - raw (bool): No parsing at all (leave ansi-to-html markers unparsed).
                - nocolor (bool): Clean out all color.
                - screenreader (bool): Use Screenreader mode.
                - send_prompt (bool): Send a prompt with parsed html

        """
        if args:
            args = list(args)
            text = args[0]
            if text is None:
                return
        else:
            return

        flags = self.protocol_flags

        options = kwargs.pop("options", {})
        raw = options.get("raw", flags.get("RAW", False))
        client_raw = options.get("client_raw", False)
        nocolor = options.get("nocolor", flags.get("NOCOLOR", False))
        screenreader = options.get("screenreader", flags.get("SCREENREADER", False))
        prompt = options.get("send_prompt", False)

        if screenreader:
            # screenreader mode cleans up output
            text = parse_ansi(text, strip_ansi=True, xterm256=False, mxp=False)
            text = _RE_SCREENREADER_REGEX.sub("", text)
        cmd = "prompt" if prompt else "text"
        if raw:
            if client_raw:
                args[0] = text
            else:
                args[0] = html.escape(text)  # escape html!
        else:
            args[0] = parse_html(text, strip_ansi=nocolor)

        # send to client on required form [cmdname, args, kwargs]
        self.sendLine(json.dumps([cmd, args, kwargs]))

    def send_prompt(self, *args, **kwargs):
        kwargs["options"].update({"send_prompt": True})
        self.send_text(*args, **kwargs)

    def send_default(self, cmdname, *args, **kwargs):
        """
        Data Evennia -> User.

        Args:
            cmdname (str): The first argument will always be the oob cmd name.
            *args (any): Remaining args will be arguments for `cmd`.

        Keyword Args:
            options (dict): These are ignored for oob commands. Use command
                arguments (which can hold dicts) to send instructions to the
                client instead.

        """
        if not cmdname == "options":
            self.sendLine(json.dumps([cmdname, args, kwargs]))



"""
Random tables - adopted from _Knave_.

"""

# Character generation tables

chargen_tables = {
    "physique": [
        "athletic",
        "brawny",
        "corpulent",
        "delicate",
        "gaunt",
        "hulking",
        "lanky",
        "ripped",
        "rugged",
        "scrawny",
        "short",
        "sinewy",
        "slender",
        "flabby",
        "statuesque",
        "stout",
        "tiny",
        "towering",
        "willowy",
        "wiry",
    ],
    "face": [
        "bloated",
        "blunt",
        "bony",
        "chiseled",
        "delicate",
        "elongated",
        "patrician",
        "pinched",
        "hawkish",
        "broken",
        "impish",
        "narrow",
        "ratlike",
        "round",
        "sunken",
        "sharp",
        "soft",
        "square",
        "wide",
        "wolfish",
    ],
    "skin": [
        "battle scar",
        "birthmark",
        "burn scar",
        "dark",
        "makeup",
        "oily",
        "pale",
        "perfect",
        "pierced",
        "pockmarked",
        "reeking",
        "tattooed",
        "rosy",
        "rough",
        "sallow",
        "sunburned",
        "tanned",
        "war paint",
        "weathered",
        "whip scar",
    ],
    "hair": [
        "bald",
        "braided",
        "bristly",
        "cropped",
        "curly",
        "disheveled",
        "dreadlocks",
        "filthy",
        "frizzy",
        "greased",
        "limp",
        "long",
        "luxurious",
        "mohawk",
        "oily",
        "ponytail",
        "silky",
        "topknot",
        "wavy",
        "wispy",
    ],
    "clothing": [
        "antique",
        "bloody",
        "ceremonial",
        "decorated",
        "eccentric",
        "elegant",
        "fashionable",
        "filthy",
        "flamboyant",
        "stained",
        "foreign",
        "frayed",
        "frumpy",
        "livery",
        "oversized",
        "patched",
        "perfumed",
        "rancid",
        "torn",
        "undersized",
    ],
    "virtue": [
        "ambitious",
        "cautious",
        "courageous",
        "courteous",
        "curious",
        "disciplined",
        "focused",
        "generous",
        "gregarious",
        "honest",
        "honorable",
        "humble",
        "idealistic",
        "just",
        "loyal",
        "merciful",
        "righteous",
        "serene",
        "stoic",
        "tolerant",
    ],
    "vice": [
        "aggressive",
        "arrogant",
        "bitter",
        "cowardly",
        "cruel",
        "deceitful",
        "flippant",
        "gluttonous",
        "greedy",
        "irascible",
        "lazy",
        "nervous",
        "prejudiced",
        "reckless",
        "rude",
        "suspicious",
        "vain",
        "vengeful",
        "wasteful",
        "whiny",
    ],
    "speech": [
        "blunt",
        "booming",
        "breathy",
        "cryptic",
        "drawling",
        "droning",
        "flowery",
        "formal",
        "gravelly",
        "hoarse",
        "mumbling",
        "precise",
        "quaint",
        "rambling",
        "rapid-fire",
        "dialect",
        "slow",
        "squeaky",
        "stuttering",
        "whispery",
    ],
    "background": [
        "alchemist",
        "beggar",
        "butcher",
        "burglar",
        "charlatan",
        "cleric",
        "cook",
        "cultist",
        "gambler",
        "herbalist",
        "magician",
        "mariner",
        "mercenary",
        "merchant",
        "outlaw",
        "performer",
        "pickpocket",
        "smuggler",
        "student",
        "tracker",
    ],
    "misfortune": [
        "abandoned",
        "addicted",
        "blackmailed",
        "condemned",
        "cursed",
        "defrauded",
        "demoted",
        "discredited",
        "disowned",
        "exiled",
        "framed",
        "haunted",
        "kidnapped",
        "mutilated",
        "poor",
        "pursued",
        "rejected",
        "replaced",
        "robbed",
        "suspected",
    ],
    "alignment": [
        ("1-5", "law"),
        ("6-15", "neutrality"),
        ("16-20", "chaos"),
    ],
    "armor": [
        ("1-3", "no armor"),
        ("4-14", "gambeson"),
        ("15-19", "brigandine"),
        ("20", "chain"),
    ],
    "helmets and shields": [
        ("1-13", "no helmet or shield"),
        ("14-16", "helmet"),
        ("17-19", "shield"),
        ("20", "helmet and shield"),
    ],
    "starting weapon": [  # note: these are all d6 dmg weapons
        ("1-7", "dagger"),
        ("8-13", "club"),
        ("14-20", "staff"),
    ],
    "dungeoning gear": [
        "rope, 50ft",
        "pulleys",
        "candles, 5",
        "chain, 10ft",
        "chalk, 10",
        "crowbar",
        "tinderbox",
        "grap. hook",
        "hammer",
        "waterskin",
        "lantern",
        "lamp oil",
        "padlock",
        "manacles",
        "mirror",
        "pole, 10ft",
        "sack",
        "tent",
        "spikes, 5",
        "torches, 5",
    ],
    "general gear 1": [
        "air bladder",
        "bear trap",
        "shovel",
        "bellows",
        "grease",
        "saw",
        "bucket",
        "caltrops",
        "chisel",
        "drill",
        "fish. rod",
        "marbles",
        "glue",
        "pick",
        "hourglass",
        "net",
        "tongs",
        "lockpicks",
        "metal file",
        "nails",
    ],
    "general gear 2": [
        "incense",
        "sponge",
        "lens",
        "perfume",
        "horn",
        "bottle",
        "soap",
        "spyglass",
        "tar pot",
        "twine",
        "fake jewels",
        "blank book",
        "card deck",
        "dice set",
        "cook pots",
        "face paint",
        "whistle",
        "instrument",
        "quill & ink",
        "small bell",
    ],
    "name": [
        "Abbo",
        "Adelaide",
        "Ellis",
        "Eleanor",
        "Lief",
        "Luanda",
        "Ablerus",
        "Agatha",
        "Eneto",
        "Elizabeth",
        "Luke",
        "Lyra",
        "Acot",
        "Aleida",
        "Enio",
        "Elspeth",
        "Martin",
        "Mabel",
        "Alexander",
        "Alexia",
        "Eral",
        "Emeline",
        "Merrick",
        "Maerwynn",
        "Almanzor",
        "Alianor",
        "Erasmus",
        "Emma",
        "Mortimer",
        "Malkyn",
        "Althalos",
        "Aline",
        "Eustace",
        "Emmony",
        "Ogden",
        "Margaret",
        "Ancelot",
        "Alma",
        "Everard",
        "Enna",
        "Oliver",
        "Margery",
        "Asher",
        "Alys",
        "Faustus",
        "Enndolynn",
        "Orion",
        "Maria",
        "Aster",
        "Amabel",
        "Favian",
        "Eve",
        "Oswald",
        "Marion",
        "Balan",
        "Amice",
        "Fendrel",
        "Evita",
        "Pelagon",
        "Matilda",
        "Balthazar",
        "Anastas",
        "Finn",
        "Felice",
        "Pello",
        "Millicent",
        "Barat",
        "Angmar",
        "Florian",
        "Fern",
        "Peyton",
        "Mirabelle",
        "Bartholomew",
        "Annabel",
        "Francis",
        "Floria",
        "Philip",
        "Muriel",
        "Basil",
        "Arabella",
        "Frederick",
        "Fredegonde",
        "Poeas",
        "Nabarne",
        "Benedict",
        "Ariana",
        "Gaidon",
        "Gillian",
        "Quinn",
        "Nell",
        "Berinon",
        "Ayleth",
        "Gavin",
        "Gloriana",
        "Ralph",
        "Nesea",
        "Bertram",
        "Barberry",
        "Geoffrey",
        "Godeleva",
        "Randolph",
        "Niree",
        "Beves",
        "Barsaba",
        "Gerard",
        "Godiva",
        "Reginald",
        "Odette",
        "Bilmer",
        "Basilia",
        "Gervase",
        "Gunnilda",
        "Reynold",
        "Odila",
        "Blanko",
        "Beatrix",
        "Gilbert",
        "Gussalen",
        "Richard",
        "Oria",
        "Bodo",
        "Benevolence",
        "Giles",
        "Gwendolynn",
        "Robert",
        "Osanna",
        "Borin",
        "Bess",
        "Godfrey",
        "Hawise",
        "Robin",
        "Ostrythe",
        "Bryce",
        "Brangian",
        "Gregory",
        "Helena",
        "Roger",
        "Ottilia",
        "Carac",
        "Brigida",
        "Gringoire",
        "Helewise",
        "Ronald",
        "Panope",
        "Caspar",
        "Brunhild",
        "Gunthar",
        "Hester",
        "Rowan",
        "Paternain",
        "Cassius",
        "Camilla",
        "Guy",
        "Hildegard",
        "Rulf",
        "Pechel",
        "Cedric",
        "Canace",
        "Gyras",
        "Idony",
        "Sabin",
        "Pepper",
        "Cephalos",
        "Cecily",
        "Hadrian",
        "Isabella",
        "Sevrin",
        "Petronilla",
        "Chadwick",
        "Cedany",
        "Hedelf",
        "Iseult",
        "Silas",
        "Phrowenia",
        "Charillos",
        "Christina",
        "Hewelin",
        "Isolde",
        "Simon",
        "Poppy",
        "Charles",
        "Claramunda",
        "Hilderith",
        "Jacquelyn",
        "Solomon",
        "Quenell",
        "Chermon",
        "Clarice",
        "Humbert",
        "Jasmine",
        "Stephen",
        "Raisa",
        "Clement",
        "Clover",
        "Hyllus",
        "Jessamine",
        "Terrowin",
        "Reyna",
        "Clifton",
        "Collette",
        "Ianto",
        "Josselyn",
        "Thomas",
        "Rixende",
        "Clovis",
        "Constance",
        "Ibykos",
        "Juliana",
        "Tristan",
        "Rosamund",
        "Cyon",
        "Damaris",
        "Inigo",
        "Karitate",
        "Tybalt",
        "Rose",
        "Dain",
        "Daphne",
        "Itylus",
        "Katelyn",
        "Ulric",
        "Ryia",
        "Dalmas",
        "Demona",
        "James",
        "Katja",
        "Walter",
        "Sarah",
        "Danor",
        "Dimia",
        "Jasper",
        "Katrina",
        "Wander",
        "Seraphina",
        "Destrian",
        "Dione",
        "Jiles",
        "Kaylein",
        "Warin",
        "Thea",
        "Domeka",
        "Dorothea",
        "Joffridus",
        "Kinna",
        "Waverly",
        "Trillby",
        "Donald",
        "Douce",
        "Jordan",
        "Krea",
        "Willahelm",
        "Wendel",
        "Doran",
        "Duraina",
        "Joris",
        "Kypris",
        "William",
        "Wilberga",
        "Dumphey",
        "Dyota",
        "Josef",
        "Landerra",
        "Wimarc",
        "Winifred",
        "Eadmund",
        "Eberhild",
        "Laurence",
        "Larraza",
        "Wystan",
        "Wofled",
        "Eckardus",
        "Edelot",
        "Leofrick",
        "Linet",
        "Xalvador",
        "Wymarc",
        "Edward",
        "Edyva",
        "Letholdus",
        "Loreena",
        "Zane",
        "Ysmay",
    ],
}


reactions = [
    ("2", "Hostile"),
    ("3-5", "Unfriendly"),
    ("6-8", "Unsure"),
    ("9-11", "Talkative"),
    ("12", "Helpful"),
]

initiative = [
    ("1-3", "Enemy acts first"),
    ("4-6", "PC acts first"),
]


death_and_dismemberment = [
    "dead",
    "dead",  # original says 'dismemberment' here, we don't simulate this
    "weakened",  # -1d4 STR
    "unsteady",  # -1d4 DEX
    "sickly",  # -1d4 CON
    "addled",  # -1d4 INT
    "rattled",  # -1d4 WIS
    "disfigured",  # -1d4 CHA
]



"""
Test mail contrib

"""

from evennia.commands.default.tests import BaseEvenniaCommandTest

from . import mail


class TestMail(BaseEvenniaCommandTest):
    def test_mail(self):
        self.call(mail.CmdMail(), "2", "'2' is not a valid mail id.", caller=self.account)
        self.call(mail.CmdMail(), "test", "'test' is not a valid mail id.", caller=self.account)
        self.call(mail.CmdMail(), "", "There are no messages in your inbox.", caller=self.account)
        self.call(
            mail.CmdMailCharacter(),
            "Char=Message 1",
            "You have received a new @mail from Char|You sent your message.",
            caller=self.char1,
        )
        self.call(
            mail.CmdMailCharacter(), "Char=Message 2", "You sent your message.", caller=self.char2
        )
        self.call(
            mail.CmdMail(),
            "TestAccount2=Message 2",
            "You have received a new @mail from TestAccount2",
            caller=self.account2,
        )
        self.call(
            mail.CmdMail(), "TestAccount=Message 1", "You sent your message.", caller=self.account2
        )
        self.call(
            mail.CmdMail(), "TestAccount=Message 2", "You sent your message.", caller=self.account2
        )
        self.call(mail.CmdMail(), "", "| ID    From              Subject", caller=self.account)
        self.call(mail.CmdMail(), "2", "From: TestAccount2", caller=self.account)
        self.call(
            mail.CmdMail(),
            "/forward TestAccount2 = 1/Forward message",
            "You sent your message.|Message forwarded.",
            caller=self.account,
        )
        self.call(
            mail.CmdMail(), "/reply 2=Reply Message2", "You sent your message.", caller=self.account
        )
        self.call(mail.CmdMail(), "/delete 2", "Message 2 deleted", caller=self.account)



# encoding=utf-8
"""
Test form
"""

FORMCHAR = "x"
TABLECHAR = "c"


FORM = """
.------------------------------------------------.
|                                                |
|  Name: xxxxxAAxxxx    Account: xxxxxxx2xxxxxxx |
|        xxxxxxxxxxx                             |
|                                                |
 >----------------------------------------------<
|                                                |
| Desc:  xxxxxxxxxxx    STR: x4x    DEX: x5x     |
|        xxxxx3xxxxx    INT: x6x    STA: x7x     |
|        xxxxxxxxxxx    LUC: x8x    MAG: x9x     |
|                                                |
 >----------.-----------------------------------<
|           |                                    |
| ccccccccc | cccccccccccccccccccccccccccccccccc |
| ccccccccc | cccccccccccccccccccccccccccccccccc |
| cccAccccc | cccccccccccccccccccccccccccccccccc |
| ccccccccc | cccccccccccccccccccccccccccccccccc |
| ccccccccc | ccccccccccccccccBccccccccccccccccc |
|           |                                    |
 -----------`-------------------------------------
 Footer: xxxFxxx
 info
"""



"""Tests for batchprocessors """

import codecs
import textwrap

import mock
from django.test import TestCase

from evennia.utils import batchprocessors, utils


class TestBatchprocessorErrors(TestCase):
    @mock.patch.object(utils, "pypath_to_realpath", return_value=[])
    def test_read_batchfile_raises_IOError(self, _):
        with self.assertRaises(IOError):
            batchprocessors.read_batchfile("foopath")

    @mock.patch.object(utils, "pypath_to_realpath", return_value=["pathfoo"])
    @mock.patch.object(codecs, "open", side_effect=ValueError("decodeerr"))
    @mock.patch.object(batchprocessors, "_ENCODINGS", ["fooencoding"])
    def test_read_batchfile_raises_UnicodeDecodeError(self, *_):
        with self.assertRaises(UnicodeDecodeError, msg="decodeerr"):
            batchprocessors.read_batchfile("foopath")


class TestBatchCommandProcessor(TestCase):
    @mock.patch.object(batchprocessors, "read_batchfile")
    def test_parses_2_commands(self, mocked_read):
        mocked_read.return_value = textwrap.dedent(
            r"""
        @create rock
        #

        @set rock/desc =
        A big rock. You can tell is ancient.
        #

        """
        )
        commands = batchprocessors.BATCHCMD.parse_file("foopath")
        self.assertEqual(
            ["@create rock", "@set rock/desc =\nA big rock. You can tell is ancient."], commands
        )

    @mock.patch.object(batchprocessors, "read_batchfile")
    def test_parses_INSERT(self, mocked_read):
        mocked_read.side_effect = [
            textwrap.dedent(
                r"""
            @create sky
            #
            #INSERT another.ev
            #
            @create sun
            #
            """
            ),
            textwrap.dedent(
                r"""
            @create bird
            #
            @create cloud
            #
            """
            ),
        ]
        commands = batchprocessors.BATCHCMD.parse_file("foopath")
        self.assertEqual(commands, ["@create sky", "@create bird", "@create cloud", "@create sun"])
        self.assertEqual(
            mocked_read.mock_calls,
            [mock.call("foopath", file_ending=".ev"), mock.call("another.ev", file_ending=".ev")],
        )

    @mock.patch.object(batchprocessors, "read_batchfile")
    def test_parses_INSERT_raises_IOError(self, mocked_read):
        mocked_read.side_effect = [
            textwrap.dedent(
                r"""
            @create sky
            #
            #INSERT x
            #
            @create sun
            #
            """
            ),
            IOError,
        ]
        with self.assertRaises(IOError, msg="#INSERT x failed."):
            batchprocessors.BATCHCMD.parse_file("foopath")
        self.assertEqual(
            mocked_read.mock_calls,
            [mock.call("foopath", file_ending=".ev"), mock.call("x", file_ending=".ev")],
        )


class TestBatchCodeProcessor(TestCase):
    @mock.patch.object(batchprocessors, "read_batchfile")
    def test_parses_one_codeblock(self, mocked_read):
        mocked_read.return_value = textwrap.dedent(
            r"""
        print("Hello")
        """
        )
        commands = batchprocessors.BATCHCODE.parse_file("foopath")
        self.assertEqual(['# batchcode code:\n\nprint("Hello")\n'], commands)

    @mock.patch.object(batchprocessors, "read_batchfile")
    def test_parses_codeblocks(self, mocked_read):
        mocked_read.return_value = textwrap.dedent(
            r"""
        #CODE
        print("Hello")
        #CODE
        a = 1
        b = [1,
        2, 3]
        """
        )
        commands = batchprocessors.BATCHCODE.parse_file("foopath")
        self.assertEqual(
            [
                "# batchcode code:\n\n",
                '# batchcode code:\n\nprint("Hello")\n',
                "# batchcode code:\n\na = 1\nb = [1,\n2, 3]\n",
            ],
            commands,
        )

    @mock.patch.object(batchprocessors, "read_batchfile")
    def test_parses_header_and_two_codeblock(self, mocked_read):
        mocked_read.return_value = textwrap.dedent(
            r"""
        #HEADER
        a = 100
        #CODE
        a += 100
        #CODE
        a += 100
        a == 100
        """
        )
        commands = batchprocessors.BATCHCODE.parse_file("foopath")
        self.assertEqual(
            [
                "# batchcode header:\n\na = 100\n\n\n# batchcode code:\n\n",
                "# batchcode header:\n\na = 100\n\n\n# batchcode code:\n\na += 100\n",
                "# batchcode header:\n\na = 100\n\n\n# batchcode code:\n\na += 100\na == 100\n",
            ],
            commands,
        )

    @mock.patch.object(batchprocessors, "read_batchfile")
    def test_parses_INSERT(self, mocked_read):
        mocked_read.side_effect = [
            textwrap.dedent(
                r"""
            a = 1
            #INSERT another.py
            """
            ),
            textwrap.dedent(
                r"""
            print("Hello")
            """
            ),
        ]
        commands = batchprocessors.BATCHCODE.parse_file("foopath")
        self.assertEqual(
            commands,
            [
                "# batchcode code:\n"
                "\n"
                "a = 1\n"
                "# batchcode insert (another.py):# batchcode code:\n"
                "\n"
                'print("Hello")\n'
                "\n"
            ],
        )
        self.assertEqual(
            mocked_read.mock_calls,
            [mock.call("foopath", file_ending=".py"), mock.call("another.py", file_ending=".py")],
        )

    @mock.patch.object(batchprocessors, "read_batchfile")
    def test_parses_INSERT_raises_IOError(self, mocked_read):
        mocked_read.side_effect = [
            textwrap.dedent(
                r"""
            #INSERT x
            """
            ),
            IOError,
        ]
        with self.assertRaises(IOError, msg="#INSERT x failed."):
            batchprocessors.BATCHCODE.parse_file("foopath")
        self.assertEqual(
            mocked_read.mock_calls,
            [mock.call("foopath", file_ending=".py"), mock.call("x", file_ending=".py")],
        )

    @mock.patch("builtins.exec")
    def test_execs_codeblock(self, mocked_exec):
        err = batchprocessors.BATCHCODE.code_exec(
            '# batchcode code:\n\nprint("Hello")\n', extra_environ={}
        )
        self.assertIsNone(err)

    @mock.patch("builtins.exec")
    def test_execs_codeblock_with_extra_environ(self, mocked_exec):
        err = batchprocessors.BATCHCODE.code_exec(
            '# batchcode code:\n\nprint("Hello")\n', extra_environ={"foo": "bar", "baz": True}
        )
        self.assertIsNone(err)

    @mock.patch("builtins.exec")
    def test_execs_codeblock_raises(self, mocked_exec):
        mocked_exec.side_effect = Exception
        err = batchprocessors.BATCHCODE.code_exec(
            '# batchcode code:\n\nprint("Hello")\nprint("Evennia")', extra_environ={}
        )
        self.assertIsNotNone(err)



"""
This module implements the ssh (Secure SHell) protocol for encrypted
connections.

This depends on a generic session module that implements the actual
login procedure of the game, tracks sessions etc.

Using standard ssh client,

"""

import os
import re

from twisted.conch.interfaces import IConchUser
from twisted.cred import checkers
from twisted.cred.portal import Portal

_SSH_IMPORT_ERROR = """
ERROR: Missing crypto library for SSH. Install it with

       pip install cryptography pyasn1 bcrypt

(On older Twisted versions you may have to do 'pip install pycrypto pyasn1' instead).

If you get a compilation error you must install a C compiler and the
SSL dev headers (On Debian-derived systems this is the gcc and libssl-dev
packages).
"""

try:
    from twisted.conch.ssh.keys import Key
except ImportError:
    raise ImportError(_SSH_IMPORT_ERROR)

from django.conf import settings
from twisted.conch import interfaces as iconch
from twisted.conch.insults import insults
from twisted.conch.manhole import Manhole, recvline
from twisted.conch.manhole_ssh import ConchFactory, TerminalRealm, _Glue
from twisted.conch.ssh import common
from twisted.conch.ssh.userauth import SSHUserAuthServer
from twisted.internet import defer, protocol
from twisted.python import components

from evennia.accounts.models import AccountDB
from evennia.utils import ansi
from evennia.utils.utils import class_from_module, to_str

_RE_N = re.compile(r"\|n$")
_RE_SCREENREADER_REGEX = re.compile(
    r"%s" % settings.SCREENREADER_REGEX_STRIP, re.DOTALL + re.MULTILINE
)
_GAME_DIR = settings.GAME_DIR
_PRIVATE_KEY_FILE = os.path.join(_GAME_DIR, "server", "ssh-private.key")
_PUBLIC_KEY_FILE = os.path.join(_GAME_DIR, "server", "ssh-public.key")
_KEY_LENGTH = 2048

CTRL_C = "\x03"
CTRL_D = "\x04"
CTRL_BACKSLASH = "\x1c"
CTRL_L = "\x0c"

_NO_AUTOGEN = f"""
Evennia could not generate SSH private- and public keys ({{err}})
Using conch default keys instead.

If this error persists, create the keys manually (using the tools for your OS)
and put them here:
    {_PRIVATE_KEY_FILE}
    {_PUBLIC_KEY_FILE}
"""

_BASE_SESSION_CLASS = class_from_module(settings.BASE_SESSION_CLASS)


# not used atm
class SSHServerFactory(protocol.ServerFactory):
    """
    This is only to name this better in logs

    """

    noisy = False

    def logPrefix(self):
        return "SSH"


class SshProtocol(Manhole, _BASE_SESSION_CLASS):
    """
    Each account connecting over ssh gets this protocol assigned to
    them.  All communication between game and account goes through
    here.

    """

    noisy = False

    def __init__(self, starttuple):
        """
        For setting up the account.  If account is not None then we'll
        login automatically.

        Args:
            starttuple (tuple): A (account, factory) tuple.

        """
        self.protocol_key = "ssh"
        self.authenticated_account = starttuple[0]
        # obs must not be called self.factory, that gets overwritten!
        self.cfactory = starttuple[1]

    def terminalSize(self, width, height):
        """
        Initialize the terminal and connect to the new session.

        Args:
            width (int): Width of terminal.
            height (int): Height of terminal.

        """
        # Clear the previous input line, redraw it at the new
        # cursor position
        self.terminal.eraseDisplay()
        self.terminal.cursorHome()
        self.width = width
        self.height = height

        # Set color defaults
        for color in ("ANSI", "XTERM256", "TRUECOLOR"):
            self.protocol_flags[color] = True

        # initialize the session
        client_address = self.getClientAddress()
        client_address = client_address.host if client_address else None
        self.init_session("ssh", client_address, self.cfactory.sessionhandler)

        # since we might have authenticated already, we might set this here.
        if self.authenticated_account:
            self.logged_in = True
            self.uid = self.authenticated_account.id
        self.sessionhandler.connect(self)

    def connectionMade(self):
        """
        This is called when the connection is first established.

        """
        recvline.HistoricRecvLine.connectionMade(self)
        self.keyHandlers[CTRL_C] = self.handle_INT
        self.keyHandlers[CTRL_D] = self.handle_EOF
        self.keyHandlers[CTRL_L] = self.handle_FF
        self.keyHandlers[CTRL_BACKSLASH] = self.handle_QUIT

        # initalize

    def handle_INT(self):
        """
        Handle ^C as an interrupt keystroke by resetting the current
        input variables to their initial state.

        """
        self.lineBuffer = []
        self.lineBufferIndex = 0

        self.terminal.nextLine()
        self.terminal.write("KeyboardInterrupt")
        self.terminal.nextLine()

    def handle_EOF(self):
        """
        Handles EOF generally used to exit.

        """
        if self.lineBuffer:
            self.terminal.write("\a")
        else:
            self.handle_QUIT()

    def handle_FF(self):
        """
        Handle a 'form feed' byte - generally used to request a screen
        refresh/redraw.

        """
        self.terminal.eraseDisplay()
        self.terminal.cursorHome()

    def handle_QUIT(self):
        """
        Quit, end, and lose the connection.

        """
        self.terminal.loseConnection()

    def connectionLost(self, reason=None):
        """
        This is executed when the connection is lost for whatever
        reason. It can also be called directly, from the disconnect
        method.

        Args:
            reason (str): Motivation for loosing connection.

        """
        insults.TerminalProtocol.connectionLost(self, reason)
        self.sessionhandler.disconnect(self)
        self.terminal.loseConnection()

    def getClientAddress(self):
        """
        Get client address.

        Returns:
            address_and_port (tuple): The client's address and port in
                a tuple. For example `('127.0.0.1', 41917)`.

        """
        return self.terminal.transport.getPeer()

    def lineReceived(self, string):
        """
        Communication User -> Evennia. Any line return indicates a
        command for the purpose of the MUD.  So we take the user input
        and pass it on to the game engine.

        Args:
            string (str): Input text.

        """
        self.sessionhandler.data_in(self, text=string)

    def sendLine(self, string):
        """
        Communication Evennia -> User.  Any string sent should
        already have been properly formatted and processed before
        reaching this point.

        Args:
            string (str): Output text.

        """
        for line in string.split("\n"):
            # the telnet-specific method for sending
            self.terminal.write(line)
            self.terminal.nextLine()

    # session-general method hooks

    def at_login(self):
        """
        Called when this session gets authenticated by the server.
        """
        pass

    def disconnect(self, reason="Connection closed. Goodbye for now."):
        """
        Disconnect from server.

        Args:
            reason (str): Motivation for disconnect.

        """
        if reason:
            self.data_out(text=((reason,), {}))
        self.connectionLost(reason)

    def data_out(self, **kwargs):
        """
        Data Evennia -> User

        Keyword Args:
            kwargs (any): Options to the protocol.

        """
        self.sessionhandler.data_out(self, **kwargs)

    def send_text(self, *args, **kwargs):
        """
        Send text data. This is an in-band telnet operation.

        Args:
            text (str): The first argument is always the text string to send. No other arguments
                are considered.
        Keyword Args:
            options (dict): Send-option flags (booleans)

                - mxp: enforce mxp link support.
                - ansi: enforce no ansi colors.
                - xterm256: enforce xterm256 colors, regardless of ttype setting.
                - nocolor: strip all colors.
                - raw: pass string through without any ansi processing
                  (i.e. include evennia ansi markers but do not
                  convert them into ansi tokens)
                - echo: turn on/off line echo on the client. turn
                  off line echo for client, for example for password.
                  note that it must be actively turned back on again!

        """
        # print "telnet.send_text", args,kwargs  # DEBUG
        text = args[0] if args else ""
        if text is None:
            return
        text = to_str(text)

        # handle arguments
        options = kwargs.get("options", {})
        flags = self.protocol_flags
        xterm256 = options.get("xterm256", flags.get("XTERM256", True))
        useansi = options.get("ansi", flags.get("ANSI", True))
        raw = options.get("raw", flags.get("RAW", False))
        nocolor = options.get("nocolor", flags.get("NOCOLOR") or not (xterm256 or useansi))
        # echo = options.get("echo", None)  # DEBUG
        screenreader = options.get("screenreader", flags.get("SCREENREADER", False))

        if screenreader:
            # screenreader mode cleans up output
            text = ansi.parse_ansi(text, strip_ansi=True, xterm256=False, mxp=False)
            text = _RE_SCREENREADER_REGEX.sub("", text)

        if raw:
            # no processing
            self.sendLine(text)
            return
        else:
            # we need to make sure to kill the color at the end in order
            # to match the webclient output.
            linetosend = ansi.parse_ansi(
                _RE_N.sub("", text) + ("||n" if text.endswith("|") else "|n"),
                strip_ansi=nocolor,
                xterm256=xterm256,
                mxp=False,
            )
            self.sendLine(linetosend)

    def send_prompt(self, *args, **kwargs):
        self.send_text(*args, **kwargs)

    def send_default(self, *args, **kwargs):
        pass


class ExtraInfoAuthServer(SSHUserAuthServer):
    noisy = False

    def auth_password(self, packet):
        """
        Password authentication.

        Used mostly for setting up the transport so we can query
        username and password later.

        Args:
            packet (Packet): Auth packet.

        """
        password = common.getNS(packet[1:])[0]
        c = checkers.UsernamePassword(self.user, password)
        c.transport = self.transport
        return self.portal.login(c, None, IConchUser).addErrback(self._ebPassword)


class AccountDBPasswordChecker(object):
    """
    Checks the django db for the correct credentials for
    username/password otherwise it returns the account or None which is
    useful for the Realm.

    """

    noisy = False
    credentialInterfaces = (checkers.IUsernamePassword,)

    def __init__(self, factory):
        """
        Initialize the factory.

        Args:
            factory (SSHFactory): Checker factory.

        """
        self.factory = factory
        super().__init__()

    def requestAvatarId(self, c):
        """
        Generic credentials.

        """
        up = checkers.IUsernamePassword(c, None)
        username = up.username
        password = up.password
        account = AccountDB.objects.get_account_from_name(username)
        res = (None, self.factory)
        if account and account.check_password(password):
            res = (account, self.factory)
        return defer.succeed(res)


class PassAvatarIdTerminalRealm(TerminalRealm):
    """
    Returns an avatar that passes the avatarId through to the
    protocol.  This is probably not the best way to do it.

    """

    noisy = False

    def _getAvatar(self, avatarId):
        comp = components.Componentized()
        user = self.userFactory(comp, avatarId)
        sess = self.sessionFactory(comp)

        sess.transportFactory = self.transportFactory
        sess.chainedProtocolFactory = lambda: self.chainedProtocolFactory(avatarId)

        comp.setComponent(iconch.IConchUser, user)
        comp.setComponent(iconch.ISession, sess)

        return user


class TerminalSessionTransport_getPeer(object):
    """
    Taken from twisted's TerminalSessionTransport which doesn't
    provide getPeer to the transport.  This one does.

    """

    noisy = False

    def __init__(self, proto, chainedProtocol, avatar, width, height):
        self.proto = proto
        self.avatar = avatar
        self.chainedProtocol = chainedProtocol

        session = self.proto.session

        self.proto.makeConnection(
            _Glue(
                write=self.chainedProtocol.dataReceived,
                loseConnection=lambda: avatar.conn.sendClose(session),
                name="SSH Proto Transport",
            )
        )

        def loseConnection():
            self.proto.loseConnection()

        def getPeer():
            return session.conn.transport.transport.getPeer()

        self.chainedProtocol.makeConnection(
            _Glue(
                getPeer=getPeer,
                write=self.proto.write,
                loseConnection=loseConnection,
                name="Chained Proto Transport",
            )
        )

        self.chainedProtocol.terminalProtocol.terminalSize(width, height)


def getKeyPair(pubkeyfile, privkeyfile):
    """
    This function looks for RSA keypair files in the current directory. If they
    do not exist, the keypair is created.
    """

    if not (os.path.exists(pubkeyfile) and os.path.exists(privkeyfile)):
        # No keypair exists. Generate a new RSA keypair
        from cryptography.hazmat.backends import default_backend
        from cryptography.hazmat.primitives.asymmetric import rsa

        rsa_key = Key(
            rsa.generate_private_key(
                public_exponent=65537, key_size=_KEY_LENGTH, backend=default_backend()
            )
        )
        public_key_string = rsa_key.public().toString(type="OPENSSH").decode()
        private_key_string = rsa_key.toString(type="OPENSSH").decode()

        # save keys for the future.
        with open(privkeyfile, "wt") as pfile:
            pfile.write(private_key_string)
            print("Created SSH private key in '{}'".format(_PRIVATE_KEY_FILE))
        with open(pubkeyfile, "wt") as pfile:
            pfile.write(public_key_string)
            print("Created SSH public key in '{}'".format(_PUBLIC_KEY_FILE))
    else:
        with open(pubkeyfile) as pfile:
            public_key_string = pfile.read()
        with open(privkeyfile) as pfile:
            private_key_string = pfile.read()

    return Key.fromString(public_key_string), Key.fromString(private_key_string)


def makeFactory(configdict):
    """
    Creates the ssh server factory.
    """

    def chainProtocolFactory(username=None):
        return insults.ServerProtocol(
            configdict["protocolFactory"],
            *configdict.get("protocolConfigdict", (username,)),
            **configdict.get("protocolKwArgs", {}),
        )

    rlm = PassAvatarIdTerminalRealm()
    rlm.transportFactory = TerminalSessionTransport_getPeer
    rlm.chainedProtocolFactory = chainProtocolFactory
    factory = ConchFactory(Portal(rlm))
    factory.sessionhandler = configdict["sessions"]

    try:
        # create/get RSA keypair
        publicKey, privateKey = getKeyPair(_PUBLIC_KEY_FILE, _PRIVATE_KEY_FILE)
        factory.publicKeys = {b"ssh-rsa": publicKey}
        factory.privateKeys = {b"ssh-rsa": privateKey}
    except Exception as err:
        print(_NO_AUTOGEN.format(err=err))

    factory.services = factory.services.copy()
    factory.services["ssh-userauth"] = ExtraInfoAuthServer

    factory.portal.registerChecker(AccountDBPasswordChecker(factory))

    return factory



"""
Wilderness contrib - titeuf87, 2017

"""

from .wilderness import WildernessExit  # noqa
from .wilderness import WildernessMapProvider  # noqa
from .wilderness import WildernessRoom  # noqa
from .wilderness import WildernessScript  # noqa
from .wilderness import create_wilderness  # noqa
from .wilderness import enter_wilderness  # noqa
from .wilderness import get_new_coordinates  # noqa



"""
Unit tests for the crafting system contrib.

"""

from unittest import mock

from django.core.exceptions import ObjectDoesNotExist
from django.test import override_settings

from evennia.commands.default.tests import BaseEvenniaCommandTest
from evennia.utils.create import create_object
from evennia.utils.test_resources import BaseEvenniaTestCase

from . import crafting, example_recipes


class TestCraftUtils(BaseEvenniaTestCase):
    """
    Test helper utils for crafting.

    """

    maxDiff = None

    @override_settings(CRAFT_RECIPE_MODULES=[])
    def test_load_recipes(self):
        """This should only load the example module now"""

        crafting._load_recipes()
        self.assertEqual(
            crafting._RECIPE_CLASSES,
            {
                "crucible steel": example_recipes.CrucibleSteelRecipe,
                "leather": example_recipes.LeatherRecipe,
                "fireball": example_recipes.FireballRecipe,
                "heal": example_recipes.HealingRecipe,
                "oak bark": example_recipes.OakBarkRecipe,
                "pig iron": example_recipes.PigIronRecipe,
                "rawhide": example_recipes.RawhideRecipe,
                "sword": example_recipes.SwordRecipe,
                "sword blade": example_recipes.SwordBladeRecipe,
                "sword guard": example_recipes.SwordGuardRecipe,
                "sword handle": example_recipes.SwordHandleRecipe,
                "sword pommel": example_recipes.SwordPommelRecipe,
            },
        )


class _TestMaterial:
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return self.name


class TestCraftingRecipeBase(BaseEvenniaTestCase):
    """
    Test the parent recipe class.
    """

    def setUp(self):
        self.crafter = mock.MagicMock()
        self.crafter.msg = mock.MagicMock()

        self.inp1 = _TestMaterial("test1")
        self.inp2 = _TestMaterial("test2")
        self.inp3 = _TestMaterial("test3")

        self.kwargs = {"kw1": 1, "kw2": 2}

        self.recipe = crafting.CraftingRecipeBase(
            self.crafter, self.inp1, self.inp2, self.inp3, **self.kwargs
        )

    def test_msg(self):
        """Test messaging to crafter"""

        self.recipe.msg("message")
        self.crafter.msg.assert_called_with("message", {"type": "crafting"})

    def test_pre_craft(self):
        """Test validating hook"""
        self.recipe.pre_craft()
        self.assertEqual(self.recipe.validated_inputs, (self.inp1, self.inp2, self.inp3))

    def test_pre_craft_fail(self):
        """Should rase error if validation fails"""
        self.recipe.allow_craft = False
        with self.assertRaises(crafting.CraftingValidationError):
            self.recipe.pre_craft()

    def test_craft_hook__succeed(self):
        """Test craft hook, the main access method."""

        expected_result = _TestMaterial("test_result")
        self.recipe.do_craft = mock.MagicMock(return_value=expected_result)

        self.assertTrue(self.recipe.allow_craft)

        result = self.recipe.craft()

        # check result
        self.assertEqual(result, expected_result)
        self.recipe.do_craft.assert_called_with(kw1=1, kw2=2)

        # since allow_reuse is False, this usage should now be turned off
        self.assertFalse(self.recipe.allow_craft)
        # trying to re-run again should fail since rerun is False
        with self.assertRaises(crafting.CraftingError):
            self.recipe.craft()

    def test_craft_hook__fail(self):
        """Test failing the call"""

        self.recipe.do_craft = mock.MagicMock(return_value=None)

        # trigger exception
        with self.assertRaises(crafting.CraftingError):
            self.recipe.craft(raise_exception=True)

        # reset and try again without exception
        self.recipe.allow_craft = True
        result = self.recipe.craft()
        self.assertEqual(result, None)


class _MockRecipe(crafting.CraftingRecipe):
    name = "testrecipe"
    tool_tags = ["tool1", "tool2"]
    consumable_tags = ["cons1", "cons2", "cons3"]
    output_prototypes = [
        {
            "key": "Result1",
            "prototype_key": "resultprot",
            "tags": [("result1", "crafting_material")],
        }
    ]


@override_settings(CRAFT_RECIPE_MODULES=[])
class TestCraftingRecipe(BaseEvenniaTestCase):
    """
    Test the CraftingRecipe class with one recipe
    """

    maxDiff = None

    def setUp(self):
        self.crafter = mock.MagicMock()
        self.crafter.msg = mock.MagicMock()

        self.tool1 = create_object(key="tool1", tags=[("tool1", "crafting_tool")], nohome=True)
        self.tool2 = create_object(key="tool2", tags=[("tool2", "crafting_tool")], nohome=True)
        self.cons1 = create_object(key="cons1", tags=[("cons1", "crafting_material")], nohome=True)
        self.cons2 = create_object(key="cons2", tags=[("cons2", "crafting_material")], nohome=True)
        self.cons3 = create_object(key="cons3", tags=[("cons3", "crafting_material")], nohome=True)

    def tearDown(self):
        try:
            self.tool1.delete()
            self.tool2.delete()
            self.cons1.delete()
            self.cons2.delete()
            self.cons3.delete()
        except ObjectDoesNotExist:
            pass

    def test_error_format(self):
        """Test the automatic error formatter"""
        recipe = _MockRecipe(
            self.crafter, self.tool1, self.tool2, self.cons1, self.cons2, self.cons3
        )

        msg = "{missing},{tools},{consumables},{inputs},{outputs}" "{i0},{i1},{o0}"
        kwargs = {
            "missing": "foo",
            "tools": ["bar", "bar2", "bar3"],
            "consumables": ["cons1", "cons2"],
        }

        expected = {
            "missing": "foo",
            "i0": "cons1",
            "i1": "cons2",
            "i2": "cons3",
            "o0": "Result1",
            "tools": "bar, bar2, and bar3",
            "consumables": "cons1 and cons2",
            "inputs": "cons1, cons2, and cons3",
            "outputs": "Result1",
        }

        result = recipe._format_message(msg, **kwargs)
        self.assertEqual(result, msg.format_map(expected))

    def test_craft__success(self):
        """Test to create a result from the recipe"""
        recipe = _MockRecipe(
            self.crafter, self.tool1, self.tool2, self.cons1, self.cons2, self.cons3
        )

        result = recipe.craft()

        self.assertEqual(result[0].key, "Result1")
        self.assertEqual(result[0].tags.all(), ["result1", "resultprot"])
        self.crafter.msg.assert_called_with(
            recipe.success_message.format(outputs="Result1"), {"type": "crafting"}
        )

        # make sure consumables are gone
        self.assertIsNone(self.cons1.pk)
        self.assertIsNone(self.cons2.pk)
        self.assertIsNone(self.cons3.pk)
        # make sure tools remain
        self.assertIsNotNone(self.tool1.pk)
        self.assertIsNotNone(self.tool2.pk)

    def test_seed__success(self):
        """Test seed helper classmethod"""

        # needed for other dbs to pass seed
        homeroom = create_object(key="HomeRoom", nohome=True)

        # call classmethod directly
        with override_settings(DEFAULT_HOME=f"#{homeroom.id}"):
            tools, consumables = _MockRecipe.seed()

        # this should be a normal successful crafting
        recipe = _MockRecipe(self.crafter, *(tools + consumables))

        result = recipe.craft()

        self.assertEqual(result[0].key, "Result1")
        self.assertEqual(result[0].tags.all(), ["result1", "resultprot"])
        self.crafter.msg.assert_called_with(
            recipe.success_message.format(outputs="Result1"), {"type": "crafting"}
        )

        # make sure consumables are gone
        for cons in consumables:
            self.assertIsNone(cons.pk)
        # make sure tools remain
        for tool in tools:
            self.assertIsNotNone(tool.pk)

    def test_craft_missing_tool__fail(self):
        """Fail craft by missing tool2"""
        recipe = _MockRecipe(self.crafter, self.tool1, self.cons1, self.cons2, self.cons3)
        result = recipe.craft()
        self.assertFalse(result)
        self.crafter.msg.assert_called_with(
            recipe.error_tool_missing_message.format(outputs="Result1", missing="tool2"),
            {"type": "crafting"},
        )

        # make sure consumables are still there
        self.assertIsNotNone(self.cons1.pk)
        self.assertIsNotNone(self.cons2.pk)
        self.assertIsNotNone(self.cons3.pk)
        # make sure tools remain
        self.assertIsNotNone(self.tool1.pk)
        self.assertIsNotNone(self.tool2.pk)

    def test_craft_missing_cons__fail(self):
        """Fail craft by missing cons3"""
        recipe = _MockRecipe(self.crafter, self.tool1, self.tool2, self.cons1, self.cons2)
        result = recipe.craft()
        self.assertFalse(result)
        self.crafter.msg.assert_called_with(
            recipe.error_consumable_missing_message.format(outputs="Result1", missing="cons3"),
            {"type": "crafting"},
        )

        # make sure consumables are still there
        self.assertIsNotNone(self.cons1.pk)
        self.assertIsNotNone(self.cons2.pk)
        self.assertIsNotNone(self.cons3.pk)
        # make sure tools remain
        self.assertIsNotNone(self.tool1.pk)
        self.assertIsNotNone(self.tool2.pk)

    def test_craft_missing_cons__always_consume__fail(self):
        """Fail craft by missing cons3, with always-consume flag"""

        cons4 = create_object(key="cons4", tags=[("cons4", "crafting_material")], nohome=True)

        recipe = _MockRecipe(self.crafter, self.tool1, self.tool2, self.cons1, self.cons2, cons4)
        recipe.consume_on_fail = True

        result = recipe.craft()

        self.assertFalse(result)
        self.crafter.msg.assert_called_with(
            recipe.error_consumable_missing_message.format(outputs="Result1", missing="cons3"),
            {"type": "crafting"},
        )

        # make sure consumables are deleted even though we failed
        self.assertIsNone(self.cons1.pk)
        self.assertIsNone(self.cons2.pk)
        # the extra should also be gone
        self.assertIsNone(cons4.pk)
        # but cons3 should be fine since it was not included
        self.assertIsNotNone(self.cons3.pk)
        # make sure tools remain as normal
        self.assertIsNotNone(self.tool1.pk)
        self.assertIsNotNone(self.tool2.pk)

    def test_craft_wrong_tool__fail(self):
        """Fail craft by including a wrong tool"""

        wrong = create_object(key="wrong", tags=[("wrongtool", "crafting_tool")], nohome=True)

        recipe = _MockRecipe(self.crafter, self.tool1, self.tool2, self.cons1, self.cons2, wrong)
        result = recipe.craft()
        self.assertFalse(result)
        self.crafter.msg.assert_called_with(
            recipe.error_tool_excess_message.format(
                outputs="Result1", excess=wrong.get_display_name(looker=self.crafter)
            ),
            {"type": "crafting"},
        )
        # make sure consumables are still there
        self.assertIsNotNone(self.cons1.pk)
        self.assertIsNotNone(self.cons2.pk)
        self.assertIsNotNone(self.cons3.pk)
        # make sure tools remain
        self.assertIsNotNone(self.tool1.pk)
        self.assertIsNotNone(self.tool2.pk)

    def test_craft_tool_excess__fail(self):
        """Fail by too many consumables"""

        # note that this is a valid tag!
        tool3 = create_object(key="tool3", tags=[("tool2", "crafting_tool")], nohome=True)

        recipe = _MockRecipe(
            self.crafter, self.tool1, self.tool2, self.cons1, self.cons2, self.cons3, tool3
        )
        result = recipe.craft()
        self.assertFalse(result)
        self.crafter.msg.assert_called_with(
            recipe.error_tool_excess_message.format(
                outputs="Result1", excess=tool3.get_display_name(looker=self.crafter)
            ),
            {"type": "crafting"},
        )

        # make sure consumables are still there
        self.assertIsNotNone(self.cons1.pk)
        self.assertIsNotNone(self.cons2.pk)
        self.assertIsNotNone(self.cons3.pk)
        # make sure tools remain
        self.assertIsNotNone(self.tool1.pk)
        self.assertIsNotNone(self.tool2.pk)
        self.assertIsNotNone(tool3.pk)

    def test_craft_cons_excess__fail(self):
        """Fail by too many consumables"""

        # note that this is a valid tag!
        cons4 = create_object(key="cons4", tags=[("cons3", "crafting_material")], nohome=True)

        recipe = _MockRecipe(
            self.crafter, self.tool1, self.tool2, self.cons1, self.cons2, self.cons3, cons4
        )
        result = recipe.craft()
        self.assertFalse(result)
        self.crafter.msg.assert_called_with(
            recipe.error_consumable_excess_message.format(
                outputs="Result1", excess=cons4.get_display_name(looker=self.crafter)
            ),
            {"type": "crafting"},
        )

        # make sure consumables are still there
        self.assertIsNotNone(self.cons1.pk)
        self.assertIsNotNone(self.cons2.pk)
        self.assertIsNotNone(self.cons3.pk)
        self.assertIsNotNone(cons4.pk)
        # make sure tools remain
        self.assertIsNotNone(self.tool1.pk)
        self.assertIsNotNone(self.tool2.pk)

    def test_craft_tool_excess__sucess(self):
        """Allow too many consumables"""

        tool3 = create_object(key="tool3", tags=[("tool2", "crafting_tool")], nohome=True)

        recipe = _MockRecipe(
            self.crafter, self.tool1, self.tool2, self.cons1, self.cons2, self.cons3, tool3
        )
        recipe.exact_tools = False
        result = recipe.craft()
        self.assertTrue(result)
        self.crafter.msg.assert_called_with(
            recipe.success_message.format(outputs="Result1"), {"type": "crafting"}
        )

        # make sure consumables are gone
        self.assertIsNone(self.cons1.pk)
        self.assertIsNone(self.cons2.pk)
        self.assertIsNone(self.cons3.pk)
        # make sure tools remain
        self.assertIsNotNone(self.tool1.pk)
        self.assertIsNotNone(self.tool2.pk)

    def test_craft_cons_excess__sucess(self):
        """Allow too many consumables"""

        cons4 = create_object(key="cons4", tags=[("cons3", "crafting_material")], nohome=True)

        recipe = _MockRecipe(
            self.crafter, self.tool1, self.tool2, self.cons1, self.cons2, self.cons3, cons4
        )
        recipe.exact_consumables = False
        result = recipe.craft()
        self.assertTrue(result)
        self.crafter.msg.assert_called_with(
            recipe.success_message.format(outputs="Result1"), {"type": "crafting"}
        )

        # make sure consumables are gone
        self.assertIsNone(self.cons1.pk)
        self.assertIsNone(self.cons2.pk)
        self.assertIsNone(self.cons3.pk)
        # make sure tools remain
        self.assertIsNotNone(self.tool1.pk)
        self.assertIsNotNone(self.tool2.pk)

    def test_craft_tool_order__fail(self):
        """Strict tool-order recipe fail"""
        recipe = _MockRecipe(
            self.crafter, self.tool2, self.tool1, self.cons1, self.cons2, self.cons3
        )
        recipe.exact_tool_order = True
        result = recipe.craft()
        self.assertFalse(result)
        self.crafter.msg.assert_called_with(
            recipe.error_tool_order_message.format(
                outputs="Result1", missing=self.tool2.get_display_name(looker=self.crafter)
            ),
            {"type": "crafting"},
        )

        # make sure consumables are still there
        self.assertIsNotNone(self.cons1.pk)
        self.assertIsNotNone(self.cons2.pk)
        self.assertIsNotNone(self.cons3.pk)
        # make sure tools remain
        self.assertIsNotNone(self.tool1.pk)
        self.assertIsNotNone(self.tool2.pk)

    def test_craft_cons_order__fail(self):
        """Strict tool-order recipe fail"""
        recipe = _MockRecipe(
            self.crafter, self.tool1, self.tool2, self.cons3, self.cons2, self.cons1
        )
        recipe.exact_consumable_order = True
        result = recipe.craft()
        self.assertFalse(result)
        self.crafter.msg.assert_called_with(
            recipe.error_consumable_order_message.format(
                outputs="Result1", missing=self.cons3.get_display_name(looker=self.crafter)
            ),
            {"type": "crafting"},
        )

        # make sure consumables are still there
        self.assertIsNotNone(self.cons1.pk)
        self.assertIsNotNone(self.cons2.pk)
        self.assertIsNotNone(self.cons3.pk)
        # make sure tools remain
        self.assertIsNotNone(self.tool1.pk)
        self.assertIsNotNone(self.tool2.pk)


class TestCraftSword(BaseEvenniaTestCase):
    """
    Test the `craft` function by crafting the example sword.

    """

    def setUp(self):
        self.crafter = mock.MagicMock()
        self.crafter.msg = mock.MagicMock()

    @override_settings(CRAFT_RECIPE_MODULES=[], DEFAULT_HOME="#999999")
    @mock.patch("evennia.contrib.game_systems.crafting.example_recipes.random")
    def test_craft_sword(self, mockrandom):
        """
        Craft example sword. For the test, every crafting works.

        """
        # make sure every craft succeeds
        mockrandom.random = mock.MagicMock(return_value=0.2)

        def _co(key, tagkey, is_tool=False):
            tagcat = "crafting_tool" if is_tool else "crafting_material"
            return create_object(key=key, tags=[(tagkey, tagcat)], nohome=True)

        def _craft(recipe_name, *inputs):
            """shortcut to shorten and return only one element"""
            result = crafting.craft(self.crafter, recipe_name, *inputs, raise_exception=True)
            return result[0] if len(result) == 1 else result

        # generate base materials
        iron_ore1 = _co("Iron ore ingot", "iron ore")
        iron_ore2 = _co("Iron ore ingot", "iron ore")
        iron_ore3 = _co("Iron ore ingot", "iron ore")

        ash1 = _co("Pile of Ash", "ash")
        ash2 = _co("Pile of Ash", "ash")
        ash3 = _co("Pile of Ash", "ash")

        sand1 = _co("Pile of sand", "sand")
        sand2 = _co("Pile of sand", "sand")
        sand3 = _co("Pile of sand", "sand")

        coal01 = _co("Pile of coal", "coal")
        coal02 = _co("Pile of coal", "coal")
        coal03 = _co("Pile of coal", "coal")
        coal04 = _co("Pile of coal", "coal")
        coal05 = _co("Pile of coal", "coal")
        coal06 = _co("Pile of coal", "coal")
        coal07 = _co("Pile of coal", "coal")
        coal08 = _co("Pile of coal", "coal")
        coal09 = _co("Pile of coal", "coal")
        coal10 = _co("Pile of coal", "coal")
        coal11 = _co("Pile of coal", "coal")
        coal12 = _co("Pile of coal", "coal")

        oak_wood = _co("Pile of oak wood", "oak wood")
        water = _co("Bucket of water", "water")
        fur = _co("Bundle of Animal fur", "fur")

        # tools
        blast_furnace = _co("Blast furnace", "blast furnace", is_tool=True)
        furnace = _co("Smithing furnace", "furnace", is_tool=True)
        crucible = _co("Smelting crucible", "crucible", is_tool=True)
        anvil = _co("Smithing anvil", "anvil", is_tool=True)
        hammer = _co("Smithing hammer", "hammer", is_tool=True)
        knife = _co("Working knife", "knife", is_tool=True)
        cauldron = _co("Cauldron", "cauldron", is_tool=True)

        # making pig iron
        inputs = [iron_ore1, coal01, coal02, blast_furnace]
        pig_iron1 = _craft("pig iron", *inputs)

        inputs = [iron_ore2, coal03, coal04, blast_furnace]
        pig_iron2 = _craft("pig iron", *inputs)

        inputs = [iron_ore3, coal05, coal06, blast_furnace]
        pig_iron3 = _craft("pig iron", *inputs)

        # making crucible steel
        inputs = [pig_iron1, ash1, sand1, coal07, coal08, crucible]
        crucible_steel1 = _craft("crucible steel", *inputs)

        inputs = [pig_iron2, ash2, sand2, coal09, coal10, crucible]
        crucible_steel2 = _craft("crucible steel", *inputs)

        inputs = [pig_iron3, ash3, sand3, coal11, coal12, crucible]
        crucible_steel3 = _craft("crucible steel", *inputs)

        # smithing
        inputs = [crucible_steel1, hammer, anvil, furnace]
        sword_blade = _craft("sword blade", *inputs)

        inputs = [crucible_steel2, hammer, anvil, furnace]
        sword_pommel = _craft("sword pommel", *inputs)

        inputs = [crucible_steel3, hammer, anvil, furnace]
        sword_guard = _craft("sword guard", *inputs)

        # stripping fur
        inputs = [fur, knife]
        rawhide = _craft("rawhide", *inputs)

        # making bark (tannin) and cleaned wood
        inputs = [oak_wood, knife]
        oak_bark, cleaned_oak_wood = _craft("oak bark", *inputs)

        # leathermaking
        inputs = [rawhide, oak_bark, water, cauldron]
        leather = _craft("leather", *inputs)

        # sword handle
        inputs = [cleaned_oak_wood, knife]
        sword_handle = _craft("sword handle", *inputs)

        # sword (order matters)
        inputs = [
            sword_blade,
            sword_guard,
            sword_pommel,
            sword_handle,
            leather,
            knife,
            hammer,
            furnace,
        ]
        sword = _craft("sword", *inputs)

        self.assertEqual(sword.key, "Sword")

        # make sure all materials and intermediaries are deleted
        self.assertIsNone(iron_ore1.pk)
        self.assertIsNone(iron_ore2.pk)
        self.assertIsNone(iron_ore3.pk)
        self.assertIsNone(ash1.pk)
        self.assertIsNone(ash2.pk)
        self.assertIsNone(ash3.pk)
        self.assertIsNone(sand1.pk)
        self.assertIsNone(sand2.pk)
        self.assertIsNone(sand3.pk)
        self.assertIsNone(coal01.pk)
        self.assertIsNone(coal02.pk)
        self.assertIsNone(coal03.pk)
        self.assertIsNone(coal04.pk)
        self.assertIsNone(coal05.pk)
        self.assertIsNone(coal06.pk)
        self.assertIsNone(coal07.pk)
        self.assertIsNone(coal08.pk)
        self.assertIsNone(coal09.pk)
        self.assertIsNone(coal10.pk)
        self.assertIsNone(coal11.pk)
        self.assertIsNone(coal12.pk)
        self.assertIsNone(oak_wood.pk)
        self.assertIsNone(water.pk)
        self.assertIsNone(fur.pk)
        self.assertIsNone(pig_iron1.pk)
        self.assertIsNone(pig_iron2.pk)
        self.assertIsNone(pig_iron3.pk)
        self.assertIsNone(crucible_steel1.pk)
        self.assertIsNone(crucible_steel2.pk)
        self.assertIsNone(crucible_steel3.pk)
        self.assertIsNone(sword_blade.pk)
        self.assertIsNone(sword_pommel.pk)
        self.assertIsNone(sword_guard.pk)
        self.assertIsNone(rawhide.pk)
        self.assertIsNone(oak_bark.pk)
        self.assertIsNone(leather.pk)
        self.assertIsNone(sword_handle.pk)

        # make sure all tools remain
        self.assertIsNotNone(blast_furnace)
        self.assertIsNotNone(furnace)
        self.assertIsNotNone(crucible)
        self.assertIsNotNone(anvil)
        self.assertIsNotNone(hammer)
        self.assertIsNotNone(knife)
        self.assertIsNotNone(cauldron)


@mock.patch("evennia.contrib.game_systems.crafting.crafting._load_recipes", new=mock.MagicMock())
@mock.patch(
    "evennia.contrib.game_systems.crafting.crafting._RECIPE_CLASSES",
    new={"testrecipe": _MockRecipe},
)
@override_settings(CRAFT_RECIPE_MODULES=[])
class TestCraftCommand(BaseEvenniaCommandTest):
    """Test the crafting command"""

    def setUp(self):
        super().setUp()

        tools, consumables = _MockRecipe.seed(
            tool_kwargs={"location": self.char1}, consumable_kwargs={"location": self.char1}
        )

    def test_craft__success(self):
        "Successfully craft using command"
        self.call(
            crafting.CmdCraft(),
            "testrecipe from cons1, cons2, cons3 using tool1, tool2",
            _MockRecipe.success_message.format(outputs="Result1"),
        )

    def test_craft__notools__failure(self):
        "Craft fail no tools"
        self.call(
            crafting.CmdCraft(),
            "testrecipe from cons1, cons2, cons3",
            _MockRecipe.error_tool_missing_message.format(outputs="Result1", missing="tool1"),
        )

    def test_craft__nocons__failure(self):
        self.call(
            crafting.CmdCraft(),
            "testrecipe using tool1, tool2",
            _MockRecipe.error_consumable_missing_message.format(outputs="Result1", missing="cons1"),
        )

    def test_craft__unknown_recipe__failure(self):
        self.call(
            crafting.CmdCraft(),
            "nonexistent from cons1, cons2, cons3 using tool1, tool2",
            "Unknown recipe 'nonexistent'",
        )



"""
Scripts for the in-game Python system.
"""

import re
import sys
import traceback
from datetime import datetime, timedelta
from queue import Queue

from django.conf import settings

from evennia import ChannelDB, DefaultObject, DefaultScript, ScriptDB, logger
from evennia.contrib.base_systems.ingame_python.callbackhandler import CallbackHandler
from evennia.contrib.base_systems.ingame_python.utils import (
    EVENTS,
    InterruptEvent,
    get_next_wait,
)
from evennia.utils.ansi import raw
from evennia.utils.create import create_channel
from evennia.utils.dbserialize import dbserialize
from evennia.utils.utils import all_from_module, delay, pypath_to_realpath

# Constants
RE_LINE_ERROR = re.compile(r'^  File "\<string\>", line (\d+)')


class EventHandler(DefaultScript):

    """
    The event handler that contains all events in a global script.

    This script shouldn't be created more than once.  It contains
    event (in a non-persistent attribute) and callbacks (in a
    persistent attribute).  The script method would help adding,
    editing and deleting these events and callbacks.

    """

    def at_script_creation(self):
        """Hook called when the script is created."""
        self.key = "event_handler"
        self.desc = "Global event handler"
        self.persistent = True

        # Permanent data to be stored
        self.db.callbacks = {}
        self.db.to_valid = []
        self.db.locked = []

        # Tasks
        self.db.tasks = {}
        self.at_server_start()

    def at_server_start(self):
        """Set up the event system when starting.

        Note that this hook is called every time the server restarts
        (including when it's reloaded).  This hook performs the following
        tasks:

        -   Create temporarily stored events.
        -   Generate locals (individual events' namespace).
        -   Load eventfuncs, including user-defined ones.
        -   Re-schedule tasks that aren't set to fire anymore.
        -   Effectively connect the handler to the main script.

        """
        self.ndb.events = {}
        for typeclass, name, variables, help_text, custom_call, custom_add in EVENTS:
            self.add_event(typeclass, name, variables, help_text, custom_call, custom_add)

        # Generate locals
        self.ndb.current_locals = {}
        self.ndb.fresh_locals = {}
        addresses = ["evennia.contrib.base_systems.ingame_python.eventfuncs"]
        addresses.extend(getattr(settings, "EVENTFUNCS_LOCATIONS", ["world.eventfuncs"]))
        for address in addresses:
            if pypath_to_realpath(address):
                self.ndb.fresh_locals.update(all_from_module(address))

        # Restart the delayed tasks
        now = datetime.now()
        for task_id, definition in tuple(self.db.tasks.items()):
            future, obj, event_name, locals = definition
            seconds = (future - now).total_seconds()
            if seconds < 0:
                seconds = 0

            delay(seconds, complete_task, task_id)

        # Place the script in the CallbackHandler
        from evennia.contrib.base_systems.ingame_python import typeclasses

        CallbackHandler.script = self
        DefaultObject.callbacks = typeclasses.EventObject.callbacks

        # Create the channel if non-existent
        try:
            self.ndb.channel = ChannelDB.objects.get(db_key="everror")
        except ChannelDB.DoesNotExist:
            self.ndb.channel = create_channel(
                "everror",
                desc="Event errors",
                locks="control:false();listen:perm(Builders);send:false()",
            )

    def get_events(self, obj):
        """
        Return a dictionary of events on this object.

        Args:
            obj (Object or typeclass): the connected object or a general typeclass.

        Returns:
            A dictionary of the object's events.

        Notes:
            Events would define what the object can have as
            callbacks.  Note, however, that chained callbacks will not
            appear in events and are handled separately.

            You can also request the events of a typeclass, not a
            connected object.  This is useful to get the global list
            of events for a typeclass that has no object yet.

        """
        events = {}
        all_events = self.ndb.events
        classes = Queue()
        if isinstance(obj, type):
            classes.put(obj)
        else:
            classes.put(type(obj))

        invalid = []
        while not classes.empty():
            typeclass = classes.get()
            typeclass_name = typeclass.__module__ + "." + typeclass.__name__
            for key, etype in all_events.get(typeclass_name, {}).items():
                if key in invalid:
                    continue
                if etype[0] is None:  # Invalidate
                    invalid.append(key)
                    continue
                if key not in events:
                    events[key] = etype

            # Look for the parent classes
            for parent in typeclass.__bases__:
                classes.put(parent)

        return events

    def get_variable(self, variable_name):
        """
        Return the variable defined in the locals.

        This can be very useful to check the value of a variable that can be modified in an event, and whose value will be used in code.  This system allows additional customization.

        Args:
            variable_name (str): the name of the variable to return.

        Returns:
            The variable if found in the locals.
            None if not found in the locals.

        Note:
            This will return the variable from the current locals.
            Keep in mind that locals are shared between events.  As
            every event is called one by one, this doesn't pose
            additional problems if you get the variable right after
            an event has been executed.  If, however, you differ,
            there's no guarantee the variable will be here or will
            mean the same thing.

        """
        return self.ndb.current_locals.get(variable_name)

    def get_callbacks(self, obj):
        """
        Return a dictionary of the object's callbacks.

        Args:
            obj (Object): the connected objects.

        Returns:
            A dictionary of the object's callbacks.

        Note:
            This method can be useful to override in some contexts,
            when several objects would share callbacks.

        """
        obj_callbacks = self.db.callbacks.get(obj, {})
        callbacks = {}
        for callback_name, callback_list in obj_callbacks.items():
            new_list = []
            for i, callback in enumerate(callback_list):
                callback = dict(callback)
                callback["obj"] = obj
                callback["name"] = callback_name
                callback["number"] = i
                new_list.append(callback)

            if new_list:
                callbacks[callback_name] = new_list

        return callbacks

    def add_callback(self, obj, callback_name, code, author=None, valid=False, parameters=""):
        """
        Add the specified callback.

        Args:
            obj (Object): the Evennia typeclassed object to be extended.
            callback_name (str): the name of the callback to add.
            code (str): the Python code associated with this callback.
            author (Character or Account, optional): the author of the callback.
            valid (bool, optional): should the callback be connected?
            parameters (str, optional): optional parameters.

        Note:
            This method doesn't check that the callback type exists.

        """
        obj_callbacks = self.db.callbacks.get(obj, {})
        if not obj_callbacks:
            self.db.callbacks[obj] = {}
            obj_callbacks = self.db.callbacks[obj]

        callbacks = obj_callbacks.get(callback_name, [])
        if not callbacks:
            obj_callbacks[callback_name] = []
            callbacks = obj_callbacks[callback_name]

        # Add the callback in the list
        callbacks.append(
            {
                "created_on": datetime.now(),
                "author": author,
                "valid": valid,
                "code": code,
                "parameters": parameters,
            }
        )

        # If not valid, set it in 'to_valid'
        if not valid:
            self.db.to_valid.append((obj, callback_name, len(callbacks) - 1))

        # Call the custom_add if needed
        custom_add = self.get_events(obj).get(callback_name, [None, None, None, None])[3]
        if custom_add:
            custom_add(obj, callback_name, len(callbacks) - 1, parameters)

        # Build the definition to return (a dictionary)
        definition = dict(callbacks[-1])
        definition["obj"] = obj
        definition["name"] = callback_name
        definition["number"] = len(callbacks) - 1
        return definition

    def edit_callback(self, obj, callback_name, number, code, author=None, valid=False):
        """
        Edit the specified callback.

        Args:
            obj (Object): the Evennia typeclassed object to be edited.
            callback_name (str): the name of the callback to edit.
            number (int): the callback number to be changed.
            code (str): the Python code associated with this callback.
            author (Character or Account, optional): the author of the callback.
            valid (bool, optional): should the callback be connected?

        Raises:
            RuntimeError if the callback is locked.

        Note:
            This method doesn't check that the callback type exists.

        """
        obj_callbacks = self.db.callbacks.get(obj, {})
        if not obj_callbacks:
            self.db.callbacks[obj] = {}
            obj_callbacks = self.db.callbacks[obj]

        callbacks = obj_callbacks.get(callback_name, [])
        if not callbacks:
            obj_callbacks[callback_name] = []
            callbacks = obj_callbacks[callback_name]

        # If locked, don't edit it
        if (obj, callback_name, number) in self.db.locked:
            raise RuntimeError("this callback is locked.")

        # Edit the callback
        callbacks[number].update(
            {"updated_on": datetime.now(), "updated_by": author, "valid": valid, "code": code}
        )

        # If not valid, set it in 'to_valid'
        if not valid and (obj, callback_name, number) not in self.db.to_valid:
            self.db.to_valid.append((obj, callback_name, number))
        elif valid and (obj, callback_name, number) in self.db.to_valid:
            self.db.to_valid.remove((obj, callback_name, number))

        # Build the definition to return (a dictionary)
        definition = dict(callbacks[number])
        definition["obj"] = obj
        definition["name"] = callback_name
        definition["number"] = number
        return definition

    def del_callback(self, obj, callback_name, number):
        """
        Delete the specified callback.

        Args:
            obj (Object): the typeclassed object containing the callback.
            callback_name (str): the name of the callback to delete.
            number (int): the number of the callback to delete.

        Raises:
            RuntimeError if the callback is locked.

        """
        obj_callbacks = self.db.callbacks.get(obj, {})
        callbacks = obj_callbacks.get(callback_name, [])

        # If locked, don't edit it
        if (obj, callback_name, number) in self.db.locked:
            raise RuntimeError("this callback is locked.")

        # Delete the callback itself
        try:
            code = callbacks[number]["code"]
        except IndexError:
            return
        else:
            logger.log_info(
                "Deleting callback {} {} of {}:\n{}".format(callback_name, number, obj, code)
            )
            del callbacks[number]

        # Change IDs of callbacks to be validated
        i = 0
        while i < len(self.db.to_valid):
            t_obj, t_callback_name, t_number = self.db.to_valid[i]
            if obj is t_obj and callback_name == t_callback_name:
                if t_number == number:
                    # Strictly equal, delete the callback
                    del self.db.to_valid[i]
                    i -= 1
                elif t_number > number:
                    # Change the ID for this callback
                    self.db.to_valid.insert(i, (t_obj, t_callback_name, t_number - 1))
                    del self.db.to_valid[i + 1]
            i += 1

        # Update locked callback
        for i, line in enumerate(self.db.locked):
            t_obj, t_callback_name, t_number = line
            if obj is t_obj and callback_name == t_callback_name:
                if number < t_number:
                    self.db.locked[i] = (t_obj, t_callback_name, t_number - 1)

        # Delete time-related callbacks associated with this object
        for script in obj.scripts.all():
            if isinstance(script, TimecallbackScript):
                if script.obj is obj and script.db.callback_name == callback_name:
                    if script.db.number == number:
                        script.stop()
                    elif script.db.number > number:
                        script.db.number -= 1

    def accept_callback(self, obj, callback_name, number):
        """
        Valid a callback.

        Args:
            obj (Object): the object containing the callback.
            callback_name (str): the name of the callback.
            number (int): the number of the callback.

        """
        obj_callbacks = self.db.callbacks.get(obj, {})
        callbacks = obj_callbacks.get(callback_name, [])

        # Accept and connect the callback
        callbacks[number].update({"valid": True})
        if (obj, callback_name, number) in self.db.to_valid:
            self.db.to_valid.remove((obj, callback_name, number))

    def call(self, obj, callback_name, *args, **kwargs):
        """
        Call the connected callbacks.

        Args:
            obj (Object): the Evennia typeclassed object.
            callback_name (str): the callback name to call.
            *args: additional variables for this callback.

        Keyword Args:
            number (int, optional): call just a specific callback.
            parameters (str, optional): call a callback with parameters.
            locals (dict, optional): a locals replacement.

        Returns:
            True to report the callback was called without interruption,
            False otherwise.

        """
        # First, look for the callback type corresponding to this name
        number = kwargs.get("number")
        parameters = kwargs.get("parameters")
        locals = kwargs.get("locals")

        # Errors should not pass silently
        allowed = ("number", "parameters", "locals")
        if any(k for k in kwargs if k not in allowed):
            raise TypeError(
                "Unknown keyword arguments were specified " "to call callbacks: {}".format(kwargs)
            )

        event = self.get_events(obj).get(callback_name)
        if locals is None and not event:
            logger.log_err(
                "The callback {} for the object {} (typeclass "
                "{}) can't be found".format(callback_name, obj, type(obj))
            )
            return False

        # Prepare the locals if necessary
        if locals is None:
            locals = self.ndb.fresh_locals.copy()
            for i, variable in enumerate(event[0]):
                try:
                    locals[variable] = args[i]
                except IndexError:
                    logger.log_trace(
                        "callback {} of {} ({}): need variable "
                        "{} in position {}".format(callback_name, obj, type(obj), variable, i)
                    )
                    return False
        else:
            locals = {key: value for key, value in locals.items()}

        callbacks = self.get_callbacks(obj).get(callback_name, [])
        if event:
            custom_call = event[2]
            if custom_call:
                callbacks = custom_call(callbacks, parameters)

        # Now execute all the valid callbacks linked at this address
        self.ndb.current_locals = locals
        for i, callback in enumerate(callbacks):
            if not callback["valid"]:
                continue

            if number is not None and callback["number"] != number:
                continue

            try:
                exec(callback["code"], locals, locals)
            except InterruptEvent:
                return False
            except Exception:
                etype, evalue, tb = sys.exc_info()
                trace = traceback.format_exception(etype, evalue, tb)
                self.handle_error(callback, trace)

        return True

    def handle_error(self, callback, trace):
        """
        Handle an error in a callback.

        Args:
            callback (dict): the callback representation.
            trace (list): the traceback containing the exception.

        Notes:
            This method can be useful to override to change the default
            handling of errors.  By default, the error message is sent to
            the character who last updated the callback, if connected.
            If not, display to the everror channel.

        """
        callback_name = callback["name"]
        number = callback["number"]
        obj = callback["obj"]
        oid = obj.id
        logger.log_err(
            "An error occurred during the callback {} of "
            "{} (#{}), number {}\n{}".format(callback_name, obj, oid, number + 1, "\n".join(trace))
        )

        # Create the error message
        line = "|runknown|n"
        lineno = "|runknown|n"
        for error in trace:
            if error.startswith('  File "<string>", line '):
                res = RE_LINE_ERROR.search(error)
                if res:
                    lineno = int(res.group(1))

                    # Try to extract the line
                    try:
                        line = raw(callback["code"].splitlines()[lineno - 1])
                    except IndexError:
                        continue
                    else:
                        break

        exc = raw(trace[-1].strip("\n").splitlines()[-1])
        err_msg = "Error in {} of {} (#{})[{}], line {}:" " {}\n{}".format(
            callback_name, obj, oid, number + 1, lineno, line, exc
        )

        # Inform the last updater if connected
        updater = callback.get("updated_by")
        if updater is None:
            updater = callback["created_by"]

        if updater and updater.sessions.all():
            updater.msg(err_msg)
        else:
            err_msg = "Error in {} of {} (#{})[{}], line {}:" " {}\n          {}".format(
                callback_name, obj, oid, number + 1, lineno, line, exc
            )
            self.ndb.channel.msg(err_msg)

    def add_event(self, typeclass, name, variables, help_text, custom_call, custom_add):
        """
        Add a new event for a defined typeclass.

        Args:
            typeclass (str): the path leading to the typeclass.
            name (str): the name of the event to add.
            variables (list of str): list of variable names for this event.
            help_text (str): the long help text of the event.
            custom_call (callable or None): the function to be called
                    when the event fires.
            custom_add (callable or None): the function to be called when
                    a callback is added.

        """
        if typeclass not in self.ndb.events:
            self.ndb.events[typeclass] = {}

        events = self.ndb.events[typeclass]
        if name not in events:
            events[name] = (variables, help_text, custom_call, custom_add)

    def set_task(self, seconds, obj, callback_name):
        """
        Set and schedule a task to run.

        Args:
            seconds (int, float): the delay in seconds from now.
            obj (Object): the typecalssed object connected to the event.
            callback_name (str): the callback's name.

        Notes:
            This method allows to schedule a "persistent" task.
            'utils.delay' is called, but a copy of the task is kept in
            the event handler, and when the script restarts (after reload),
            the differed delay is called again.
            The dictionary of locals is frozen and will be available
            again when the task runs.  This feature, however, is limited
            by the database: all data cannot be saved.  Lambda functions,
            class methods, objects inside an instance and so on will
            not be kept in the locals dictionary.

        """
        now = datetime.now()
        delta = timedelta(seconds=seconds)

        # Choose a free task_id
        used_ids = list(self.db.tasks.keys())
        task_id = 1
        while task_id in used_ids:
            task_id += 1

        # Collect and freeze current locals
        locals = {}
        for key, value in self.ndb.current_locals.items():
            try:
                dbserialize(value)
            except TypeError:
                continue
            else:
                locals[key] = value

        self.db.tasks[task_id] = (now + delta, obj, callback_name, locals)
        delay(seconds, complete_task, task_id)


# Script to call time-related events
class TimeEventScript(DefaultScript):

    """Gametime-sensitive script."""

    def at_script_creation(self):
        """The script is created."""
        self.start_delay = True
        self.persistent = True

        # Script attributes
        self.db.time_format = None
        self.db.event_name = "time"
        self.db.number = None

    def at_repeat(self):
        """
        Call the event and reset interval.

        It is necessary to restart the script to reset its interval
        only twice after a reload.  When the script has undergone
        down time, there's usually a slight shift in game time.  Once
        the script restarts once, it will set the average time it
        needs for all its future intervals and should not need to be
        restarted.  In short, a script that is created shouldn't need
        to restart more than once, and a script that is reloaded should
        restart only twice.

        """
        if self.db.time_format:
            # If the 'usual' time is set, use it
            seconds = self.ndb.usual
            if seconds is None:
                seconds, usual, details = get_next_wait(self.db.time_format)
                self.ndb.usual = usual

            if self.interval != seconds:
                self.restart(interval=seconds)

        if self.db.event_name and self.db.number is not None:
            obj = self.obj
            if not obj.callbacks:
                return

            event_name = self.db.event_name
            number = self.db.number
            obj.callbacks.call(event_name, obj, number=number)


# Functions to manipulate tasks
def complete_task(task_id):
    """
    Mark the task in the event handler as complete.

    Args:
        task_id (int): the task ID.

    Note:
        This function should be called automatically for individual tasks.

    """
    try:
        script = ScriptDB.objects.get(db_key="event_handler")
    except ScriptDB.DoesNotExist:
        logger.log_trace("Can't get the event handler.")
        return

    if task_id not in script.db.tasks:
        logger.log_err("The task #{} was scheduled, but it cannot be " "found".format(task_id))
        return

    delta, obj, callback_name, locals = script.db.tasks.pop(task_id)
    script.call(obj, callback_name, locals=locals)



"""
Script that saves memory and idmapper data over time.

Data will be saved to game/logs/memoryusage.log. Note that
the script will append to this file if it already exists.

Call this module directly to plot the log (requires matplotlib and numpy).
"""

import os
import sys
import time

# TODO!
# sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))
# os.environ['DJANGO_SETTINGS_MODULE'] = 'game.settings'
import evennia
from evennia.utils.idmapper import models as _idmapper

LOGFILE = "logs/memoryusage.log"
INTERVAL = 30  # log every 30 seconds


class Memplot(evennia.DefaultScript):
    """
    Describes a memory plotting action.

    """

    def at_script_creation(self):
        "Called at script creation"
        self.key = "memplot"
        self.desc = "Save server memory stats to file"
        self.start_delay = False
        self.persistent = True
        self.interval = INTERVAL
        self.db.starttime = time.time()

    def at_repeat(self):
        "Regularly save memory statistics."
        pid = os.getpid()
        rmem = (
            float(os.popen("ps -p %d -o %s | tail -1" % (pid, "rss")).read()) / 1000.0
        )  # resident memory
        vmem = (
            float(os.popen("ps -p %d -o %s | tail -1" % (pid, "vsz")).read()) / 1000.0
        )  # virtual memory
        total_num, cachedict = _idmapper.cache_size()
        t0 = (time.time() - self.db.starttime) / 60.0  # save in minutes

        with open(LOGFILE, "a") as f:
            f.write("%s, %s, %s, %s\n" % (t0, rmem, vmem, int(total_num)))


if __name__ == "__main__":
    # plot output from the file

    import numpy
    from matplotlib import pyplot as pp

    data = numpy.genfromtxt("../../../game/" + LOGFILE, delimiter=",")
    secs = data[:, 0]
    rmem = data[:, 1]
    vmem = data[:, 2]
    nobj = data[:, 3]

    # calculate derivative of obj creation
    # oderiv = (0.5*(nobj[2:] - nobj[:-2]) / (secs[2:] - secs[:-2])).copy()
    # oderiv = (0.5*(rmem[2:] - rmem[:-2]) / (secs[2:] - secs[:-2])).copy()

    fig = pp.figure()
    ax1 = fig.add_subplot(111)
    ax1.set_title("1000 bots (normal accounts with light building)")
    ax1.set_xlabel("Time (mins)")
    ax1.set_ylabel("Memory usage (MB)")
    ax1.plot(secs, rmem, "r", label="RMEM", lw=2)
    ax1.plot(secs, vmem, "b", label="VMEM", lw=2)
    ax1.legend(loc="upper left")

    ax2 = ax1.twinx()
    ax2.plot(secs, nobj, "g--", label="objs in cache", lw=2)
    # ax2.plot(secs[:-2], oderiv/60.0, "g--", label="Objs/second", lw=2)
    # ax2.plot(secs[:-2], oderiv, "g--", label="Objs/second", lw=2)
    ax2.set_ylabel("Number of objects")
    ax2.legend(loc="lower right")
    ax2.annotate("First 500 bots\nconnecting", xy=(10, 4000))
    ax2.annotate("Next 500 bots\nconnecting", xy=(350, 10000))
    # ax2.annotate("@reload", xy=(185,600))

    #    # plot mem vs cachesize
    #    nobj, rmem, vmem = nobj[:262].copy(), rmem[:262].copy(), vmem[:262].copy()
    #
    #    fig = pp.figure()
    #    ax1 = fig.add_subplot(111)
    #    ax1.set_title("Memory usage per cache size")
    #    ax1.set_xlabel("Cache size (number of objects)")
    #    ax1.set_ylabel("Memory usage (MB)")
    #    ax1.plot(nobj, rmem, "r", label="RMEM", lw=2)
    #    ax1.plot(nobj, vmem, "b", label="VMEM", lw=2)
    #

    # empirical estimate of memory usage: rmem = 35.0 + 0.0157 * Ncache
    # Ncache = int((rmem - 35.0) / 0.0157)  (rmem in MB)
    #
    #    rderiv_aver = 0.0157
    #    fig = pp.figure()
    #    ax1 = fig.add_subplot(111)
    #    ax1.set_title("Relation between memory and cache size")
    #    ax1.set_xlabel("Memory usage (MB)")
    #    ax1.set_ylabel("Idmapper Cache Size (number of objects)")
    #    rmem = numpy.linspace(35, 2000, 2000)
    #    nobjs = numpy.array([int((mem - 35.0) / 0.0157) for mem in rmem])
    #    ax1.plot(rmem, nobjs, "r", lw=2)

    pp.show()



from django import forms
from django.conf import settings
from django.contrib.auth.forms import UserCreationForm, UsernameField
from django.forms import ModelForm
from django.utils.html import escape

from evennia.utils import class_from_module


class EvenniaForm(forms.Form):
    """
    This is a stock Django form, but modified so that all values provided
    through it are escaped (sanitized). Validation is performed by the fields
    you define in the form.

    This has little to do with Evennia itself and is more general web security-
    related.

    https://www.owasp.org/index.php/Input_Validation_Cheat_Sheet#Goals_of_Input_Validation

    """

    def clean(self):
        """
        Django hook. Performed on form submission.

        Returns:
            cleaned (dict): Dictionary of key:value pairs submitted on the form.

        """
        # Call parent function
        cleaned = super().clean()

        # Escape all values provided by user
        cleaned = {k: escape(v) for k, v in cleaned.items()}
        return cleaned


class AccountForm(UserCreationForm):
    """
    This is a generic Django form tailored to the Account model.

    In this incarnation it does not allow getting/setting of attributes, only
    core User model fields (username, email, password).

    """

    class Meta:
        """
        This is a Django construct that provides additional configuration to
        the form.

        """

        # The model/typeclass this form creates
        model = class_from_module(
            settings.BASE_ACCOUNT_TYPECLASS, fallback=settings.FALLBACK_ACCOUNT_TYPECLASS
        )

        # The fields to display on the form, in the given order
        fields = ("username", "email")

        # Any overrides of field classes
        field_classes = {"username": UsernameField}

    # Username is collected as part of the core UserCreationForm, so we just need
    # to add a field to (optionally) capture email.
    email = forms.EmailField(
        help_text="A valid email address. Optional; used for password resets.", required=False
    )


class ObjectForm(EvenniaForm, ModelForm):
    """
    This is a Django form for generic Evennia Objects that allows modification
    of attributes when called from a descendent of ObjectUpdate or ObjectCreate
    views.

    It defines no fields by default; you have to do that by extending this class
    and defining what fields you want to be recorded. See the CharacterForm for
    a simple example of how to do this.

    """

    class Meta:
        """
        This is a Django construct that provides additional configuration to
        the form.

        """

        # The model/typeclass this form creates
        model = class_from_module(
            settings.BASE_OBJECT_TYPECLASS, fallback=settings.FALLBACK_OBJECT_TYPECLASS
        )

        # The fields to display on the form, in the given order
        fields = ("db_key",)

        # This lets us rename ugly db-specific keys to something more human
        labels = {"db_key": "Name"}


class CharacterForm(ObjectForm):
    """
    This is a Django form for Evennia Character objects.

    Since Evennia characters only have one attribute by default, this form only
    defines a field for that single attribute. The names of fields you define should
    correspond to their names as stored in the dbhandler; you can display
    'prettier' versions of the fieldname on the form using the 'label' kwarg.

    The basic field types are CharFields and IntegerFields, which let you enter
    text and numbers respectively. IntegerFields have some neat validation tricks
    they can do, like mandating values fall within a certain range.

    For example, a complete "age" field (which stores its value to
    `character.db.age` might look like:

    age = forms.IntegerField(
        label="Your Age",
        min_value=18, max_value=9000,
        help_text="Years since your birth.")

    Default input fields are generic single-line text boxes. You can control what
    sort of input field users will see by specifying a "widget." An example of
    this is used for the 'desc' field to show a Textarea box instead of a Textbox.

    For help in building out your form, please see:
    https://docs.djangoproject.com/en/4.1/topics/forms/#building-a-form-in-django

    For more information on fields and their capabilities, see:
    https://docs.djangoproject.com/en/4.1/ref/forms/fields/

    For more on widgets, see:
    https://docs.djangoproject.com/en/4.1/ref/forms/widgets/

    """

    class Meta:
        """
        This is a Django construct that provides additional configuration to
        the form.

        """

        # Get the correct object model
        model = class_from_module(
            settings.BASE_CHARACTER_TYPECLASS, fallback=settings.FALLBACK_CHARACTER_TYPECLASS
        )

        # Allow entry of the 'key' field
        fields = ("db_key",)

        # Rename 'key' to something more intelligible
        labels = {"db_key": "Name"}

    # Fields pertaining to configurable attributes on the Character object.
    desc = forms.CharField(
        label="Description",
        max_length=2048,
        required=False,
        widget=forms.Textarea(attrs={"rows": 3}),
        help_text="A brief description of your character.",
    )


class CharacterUpdateForm(CharacterForm):
    """
    This is a Django form for updating Evennia Character objects.

    By default it is the same as the CharacterForm, but if there are circumstances
    in which you don't want to let players edit all the same attributes they had
    access to during creation, you can redefine this form with those fields you do
    wish to allow.

    """

    pass



"""
Implements Discord chat channel integration.

The Discord API uses a mix of websockets and REST API endpoints.

In order for this integration to work, you need to have your own
discord bot set up via https://discord.com/developers/applications
with the MESSAGE CONTENT toggle switched on, and your bot token
added to `server/conf/secret_settings.py` as your  DISCORD_BOT_TOKEN
"""
import json
import os
from io import BytesIO
from random import random

from autobahn.twisted.websocket import (
    WebSocketClientFactory,
    WebSocketClientProtocol,
    connectWS,
)
from django.conf import settings
from twisted.internet import protocol, reactor, ssl, task
from twisted.web.client import Agent, FileBodyProducer, HTTPConnectionPool, readBody
from twisted.web.http_headers import Headers

from evennia.server.session import Session
from evennia.utils import class_from_module, get_evennia_version, logger
from evennia.utils.utils import delay

_BASE_SESSION_CLASS = class_from_module(settings.BASE_SESSION_CLASS)

DISCORD_API_VERSION = 10
# include version number to prevent automatically updating to breaking changes
DISCORD_API_BASE_URL = f"https://discord.com/api/v{DISCORD_API_VERSION}"

DISCORD_USER_AGENT = f"Evennia (https://www.evennia.com, {get_evennia_version(mode='short')})"
DISCORD_BOT_TOKEN = settings.DISCORD_BOT_TOKEN
DISCORD_BOT_INTENTS = settings.DISCORD_BOT_INTENTS

# Discord OP codes, alphabetic
OP_DISPATCH = 0
OP_HEARTBEAT = 1
OP_HEARTBEAT_ACK = 11
OP_HELLO = 10
OP_IDENTIFY = 2
OP_INVALID_SESSION = 9
OP_RECONNECT = 7
OP_RESUME = 6


# create quiet HTTP pool to muffle GET/POST requests
class QuietConnectionPool(HTTPConnectionPool):
    """
    A quiet version of the HTTPConnectionPool which sets the factory's
    `noisy` property to False to muffle log output.
    """

    def __init__(self, reactor, persistent=True):
        super().__init__(reactor, persistent)
        self._factory.noisy = False


_AGENT = Agent(reactor, pool=QuietConnectionPool(reactor))


def should_retry(status_code):
    """
    Helper function to check if the request should be retried later.

    Args:
        status_code (int) - The HTTP status code

    Returns:
        retry (bool) - True if request should be retried False otherwise
    """
    if status_code >= 500 and status_code <= 504:
        # these are common server error codes when the server is temporarily malfunctioning
        # in these cases, we should retry
        return True
    else:
        # handle all other cases; this can be expanded later if needed for special cases
        return False


class DiscordWebsocketServerFactory(WebSocketClientFactory, protocol.ReconnectingClientFactory):
    """
    A customized websocket client factory that navigates the Discord gateway process.

    """

    initialDelay = 1
    factor = 1.5
    maxDelay = 60
    noisy = False
    gateway = None
    resume_url = None
    is_connecting = False

    def __init__(self, sessionhandler, *args, **kwargs):
        self.uid = kwargs.get("uid")
        self.sessionhandler = sessionhandler
        self.port = None
        self.bot = None

    def get_gateway_url(self, *args, **kwargs):
        # get the websocket gateway URL from Discord
        d = _AGENT.request(
            b"GET",
            f"{DISCORD_API_BASE_URL}/gateway".encode("utf-8"),
            Headers(
                {
                    "User-Agent": [DISCORD_USER_AGENT],
                    "Authorization": [f"Bot {DISCORD_BOT_TOKEN}"],
                    "Content-Type": ["application/json"],
                }
            ),
            None,
        )

        def cbResponse(response):
            if response.code == 200:
                d = readBody(response)
                d.addCallback(self.websocket_init, *args, **kwargs)
                return d
            else:
                logger.log_warn("Discord gateway request failed.")

        d.addCallback(cbResponse)

    def websocket_init(self, payload, *args, **kwargs):
        """
        callback for when the URL is gotten
        """
        data = json.loads(str(payload, "utf-8"))
        self.is_connecting = False
        if url := data.get("url"):
            self.gateway = f"{url}/?v={DISCORD_API_VERSION}&encoding=json".encode("utf-8")
            useragent = kwargs.pop("useragent", DISCORD_USER_AGENT)
            headers = kwargs.pop(
                "headers",
                {
                    "Authorization": [f"Bot {DISCORD_BOT_TOKEN}"],
                    "Content-Type": ["application/json"],
                },
            )

            logger.log_info("Connecting to Discord Gateway...")
            WebSocketClientFactory.__init__(
                self, url, *args, headers=headers, useragent=useragent, **kwargs
            )
            self.start()
        else:
            logger.log_err("Discord did not return a websocket URL; connection cancelled.")

    def buildProtocol(self, addr):
        """
        Build new instance of protocol

        Args:
            addr (str): Not used, using factory/settings data

        """
        if hasattr(settings, "DISCORD_SESSION_CLASS"):
            protocol_class = class_from_module(
                settings.DISCORD_SESSION_CLASS, fallback=DiscordClient
            )
            protocol = protocol_class()
        else:
            protocol = DiscordClient()

        protocol.factory = self
        protocol.sessionhandler = self.sessionhandler
        return protocol

    def startedConnecting(self, connector):
        """
        Tracks reconnections for debugging.

        Args:
            connector (Connector): Represents the connection.

        """
        logger.log_info("Connecting to Discord...")

    def reconnect(self):
        """
        Force a reconnection of the bot protocol. This requires
        de-registering the session and then reattaching a new one.

        """
        # set up the reconnection
        if self.resume_url:
            self.url = self.resume_url
        elif self.gateway:
            self.url = self.gateway
        else:
            # we don't know where to reconnect to! we'll start from the beginning
            self.url = None
        # reset the internal delay, since this is a deliberate disconnect
        self.delay = self.initialDelay
        # disconnect to allow the reconnection process to kick in
        self.bot.sendClose()
        self.sessionhandler.server_disconnect(self.bot)

    def start(self):
        "Connect protocol to remote server"

        if not self.gateway:
            # we don't know where to connect to
            # get the gateway URL from Discord
            self.is_connecting = True
            self.get_gateway_url()
        elif not self.is_connecting:
            # everything is good, connect
            connectWS(self)


class DiscordClient(WebSocketClientProtocol, _BASE_SESSION_CLASS):
    """
    Implements the Discord client
    """

    nextHeartbeatCall = None
    pending_heartbeat = False
    heartbeat_interval = None
    last_sequence = 0
    session_id = None
    discord_id = None

    def __init__(self):
        WebSocketClientProtocol.__init__(self)
        _BASE_SESSION_CLASS.__init__(self)

    def at_login(self):
        pass

    def onOpen(self):
        """
        Called when connection is established.

        """
        logger.log_msg("Discord connection established.")
        self.factory.bot = self

        self.init_session("discord", "discord.gg", self.factory.sessionhandler)
        self.uid = int(self.factory.uid)
        self.logged_in = True
        self.sessionhandler.connect(self)

    def onMessage(self, payload, isBinary):
        """
        Callback fired when a complete WebSocket message was received.

        Args:
            payload (bytes): The WebSocket message received.
            isBinary (bool): Flag indicating whether payload is binary or
                             UTF-8 encoded text.

        """
        if isBinary:
            logger.log_info("DISCORD: got a binary payload for some reason")
            return
        data = json.loads(str(payload, "utf-8"))
        if seqid := data.get("s"):
            self.last_sequence = seqid

        # not sure if that error json format is for websockets, so
        # check for it just in case
        if "errors" in data:
            self.handle_error(data)
            return

        # check for discord gateway API op codes first
        if data["op"] == OP_HELLO:
            self.interval = data["d"]["heartbeat_interval"] / 1000  # convert millisec to seconds
            if self.nextHeartbeatCall:
                self.nextHeartbeatCall.cancel()
            self.nextHeartbeatCall = self.factory._batched_timer.call_later(
                self.interval * random(),
                self.doHeartbeat,
            )
            if self.session_id:
                # we already have a session; try to resume instead
                self.resume()
            else:
                self.identify()
        elif data["op"] == OP_HEARTBEAT_ACK:
            # our last heartbeat was acknowledged, so reset the "pending" flag
            self.pending_heartbeat = False
        elif data["op"] == OP_HEARTBEAT:
            # Discord wants us to send a heartbeat immediately
            self.doHeartbeat(force=True)
        elif data["op"] == OP_INVALID_SESSION:
            # Discord doesn't like our current session; reconnect for a new one
            logger.log_msg("Discord: received 'Invalid Session' opcode. Reconnecting.")
            if data["d"] == False:
                # can't resume, clear existing resume data
                self.session_id = None
                self.factory.resume_url = None
            self.factory.reconnect()
        elif data["op"] == OP_RECONNECT:
            # reconnect as requested; Discord does this regularly for server load balancing
            logger.log_msg("Discord: received 'Reconnect' opcode. Reconnecting.")
            self.factory.reconnect()
        elif data["op"] == OP_DISPATCH:
            # handle the general dispatch opcode events by type
            if data["t"] == "READY":
                # our recent identification is valid; process new session info
                self.connection_ready(data["d"])
            else:
                # general message, pass on to data_in
                self.data_in(data=data)

    def onClose(self, wasClean, code=None, reason=None):
        """
        This is executed when the connection is lost for whatever
        reason. it can also be called directly, from the disconnect
        method.

        Args:
            wasClean (bool): ``True`` if the WebSocket was closed cleanly.
            code (int or None): Close status as sent by the WebSocket peer.
            reason (str or None): Close reason as sent by the WebSocket peer.

        """
        self.sessionhandler.disconnect(self)
        if self.nextHeartbeatCall:
            self.nextHeartbeatCall.cancel()
            self.nextHeartbeatCall = None
        if wasClean:
            logger.log_info(f"Discord connection closed ({code}) reason: {reason}")
        else:
            logger.log_info(f"Discord connection lost.")

    def _send_json(self, data):
        """
        Post JSON data to the websocket

        Args:
            data (dict): content to send.

        """
        return self.sendMessage(json.dumps(data).encode("utf-8"))

    def _post_json(self, url, data, **kwargs):
        """
        Post JSON data to a REST API endpoint

        Args:
            url (str) - The API path which is being posted to
            data (dict) - Content to be sent
        """
        url = f"{DISCORD_API_BASE_URL}/{url}"
        body = FileBodyProducer(BytesIO(json.dumps(data).encode("utf-8")))
        request_type = kwargs.pop("type", "POST")

        d = _AGENT.request(
            request_type.encode("utf-8"),
            url.encode("utf-8"),
            Headers(
                {
                    "User-Agent": [DISCORD_USER_AGENT],
                    "Authorization": [f"Bot {DISCORD_BOT_TOKEN}"],
                    "Content-Type": ["application/json"],
                }
            ),
            body,
        )

        def cbResponse(response):
            if response.code == 200 or response.code == 204:
                d = readBody(response)
                d.addCallback(self.post_response)
                return d
            elif should_retry(response.code):
                delay(300, self._post_json, url, data, **kwargs)

        d.addCallback(cbResponse)

    def post_response(self, body, **kwargs):
        """
        Process the response from sending a POST request

        Args:
            body (bytes) - The post response body
        """
        data = json.loads(body)
        if "errors" in data:
            self.handle_error(data)

    def handle_error(self, data, **kwargs):
        """
        General hook for processing errors.

        Args:
            data (dict) - The received error data

        """
        logger.log_err(str(data))

    def resume(self):
        """
        Called after a reconnection to re-identify and replay missed events

        """
        if not self.last_sequence or not self.session_id:
            # we have no known state to resume from, identify normally
            self.identify()

        # build a RESUME request for Discord and send it
        data = {
            "op": OP_RESUME,
            "d": {
                "token": DISCORD_BOT_TOKEN,
                "session_id": self.session_id,
                "s": self.sequence_id,
            },
        }
        self._send_json(data)

    def disconnect(self, reason=None):
        """
        Generic hook for the engine to call in order to
        disconnect this protocol.

        Args:
            reason (str or None): Motivation for the disconnection.

        """
        self.sendClose(self.CLOSE_STATUS_CODE_NORMAL, reason)

    def identify(self, *args, **kwargs):
        """
        Send Discord authentication. This should be sent once heartbeats begin.

        """
        data = {
            "op": 2,
            "d": {
                "token": DISCORD_BOT_TOKEN,
                "intents": DISCORD_BOT_INTENTS,
                "properties": {
                    "os": os.name,
                    "browser": DISCORD_USER_AGENT,
                    "device": DISCORD_USER_AGENT,
                },
            },
        }
        self._send_json(data)

    def connection_ready(self, data):
        """
        Process READY data for relevant bot info.
        """
        self.factory.resume_url = data["resume_gateway_url"]
        self.session_id = data["session_id"]
        self.discord_id = data["user"]["id"]

    def doHeartbeat(self, *args, **kwargs):
        """
        Send heartbeat to Discord.

        """
        if not self.pending_heartbeat or kwargs.get("force"):
            if self.nextHeartbeatCall:
                self.nextHeartbeatCall.cancel()
            # send the heartbeat
            data = {"op": 1, "d": self.last_sequence}
            self._send_json(data)
            # track that we sent a heartbeat, in case we don't receive an ACK
            self.pending_heartbeat = True
            self.nextHeartbeatCall = self.factory._batched_timer.call_later(
                self.interval,
                self.doHeartbeat,
            )
        else:
            # we didn't get a response since the last heartbeat; reconnect
            self.factory.reconnect()

    def send_channel(self, text, channel_id, **kwargs):
        """
        Send a message from an Evennia channel to a Discord channel.

        Use with session.msg(channel=(message, channel, sender))

        """

        data = {"content": text}
        data.update(kwargs)
        self._post_json(f"channels/{channel_id}/messages", data)

    def send_nickname(self, text, guild_id, user_id, **kwargs):
        """
        Changes a user's nickname on a Discord server.

        Use with session.msg(nickname=(new_nickname, guild_id, user_id))
        """

        data = {"nick": text}
        data.update(kwargs)
        self._post_json(f"guilds/{guild_id}/members/{user_id}", data, type="PATCH")

    def send_role(self, role_id, guild_id, user_id, **kwargs):
        data = kwargs
        self._post_json(f"guilds/{guild_id}/members/{user_id}/roles/{role_id}", data, type="PUT")

    def send_default(self, *args, **kwargs):
        """
        Ignore other outputfuncs

        """
        pass

    def data_in(self, data, **kwargs):
        """
        Process incoming data from Discord and sent to the Evennia server

        Args:
            data (dict): Converted json data.

        """
        action_type = data.get("t", "UNKNOWN")

        if action_type == "MESSAGE_CREATE":
            # someone posted a message on Discord that the bot can see
            data = data["d"]
            if data["author"]["id"] == self.discord_id:
                # it's by the bot itself! disregard
                return
            message = data["content"]
            channel_id = data["channel_id"]
            keywords = {"channel_id": channel_id}
            if "guild_id" in data:
                # message received to a Discord channel
                keywords["type"] = "channel"
                author = data["member"]["nick"] or data["author"]["username"]
                author_id = data["author"]["id"]
                keywords["sender"] = (author_id, author)
                keywords["guild_id"] = data["guild_id"]

            else:
                # message sent directly to the bot account via DM
                keywords["type"] = "direct"
                author = data["author"]["username"]
                author_id = data["author"]["id"]
                keywords["sender"] = (author_id, author)

            # pass the processed data to the server
            self.sessionhandler.data_in(self, bot_data_in=(message, keywords))

        elif action_type in ("GUILD_CREATE", "GUILD_UPDATE"):
            # we received the current status of a guild the bot is on; process relevant info
            data = data["d"]
            keywords = {"type": "guild", "guild_id": data["id"], "guild_name": data["name"]}
            keywords["channels"] = {
                chan["id"]: {"name": chan["name"], "guild": data["name"]}
                for chan in data["channels"]
                if chan["type"] == 0
            }
            # send the possibly-updated guild and channel data to the server
            self.sessionhandler.data_in(self, bot_data_in=("", keywords))

        elif "DELETE" in action_type:
            # deletes should possibly be handled separately to check for channel removal
            # for now, just ignore
            pass

        else:
            # send the data for any other action types on to the bot as-is for optional server-side handling
            keywords = {"type": action_type}
            keywords.update(data["d"])
            self.sessionhandler.data_in(self, bot_data_in=("", keywords))



from django.contrib.auth import authenticate, login

from evennia.accounts.models import AccountDB
from evennia.utils import ip_from_request, logger


class OriginIpMiddleware:
    """
    This Django Middleware simply sets the request.origin_ip attribute to what is
    respected by the Evennia Server, taking into account settings.UPSTREAM_IPS.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        request.origin_ip = ip_from_request(request)
        return self.get_response(request)


class SharedLoginMiddleware(object):
    """
    Handle the shared login between website and webclient.

    """

    def __init__(self, get_response):
        # One-time configuration and initialization.
        self.get_response = get_response

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.

        # Synchronize credentials between webclient and website
        # Must be performed *before* rendering the view (issue #1723)
        self.make_shared_login(request)

        # Process view
        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.

        # Return processed view
        return response

    @classmethod
    def make_shared_login(cls, request):
        csession = request.session
        account = request.user
        website_uid = csession.get("website_authenticated_uid", None)
        webclient_uid = csession.get("webclient_authenticated_uid", None)

        if not csession.session_key:
            # this is necessary to build the sessid key
            csession.save()

        if account.is_authenticated:
            # Logged into website
            if website_uid is None:
                # fresh website login (just from login page)
                csession["website_authenticated_uid"] = account.id
            if webclient_uid is None:
                # auto-login web client
                csession["webclient_authenticated_uid"] = account.id

        elif webclient_uid:
            # Not logged into website, but logged into webclient
            if website_uid is None:
                csession["website_authenticated_uid"] = account.id
                account = AccountDB.objects.get(id=webclient_uid)
                try:
                    # calls our custom authenticate, in web/utils/backend.py
                    authenticate(autologin=account)
                    login(request, account)
                except AttributeError:
                    logger.log_trace()

        if csession.get("webclient_authenticated_uid", None):
            # set a nonce to prevent the webclient from erasing the webclient_authenticated_uid value
            csession["webclient_authenticated_nonce"] = (
                csession.get("webclient_authenticated_nonce", 0) + 1
            )
            # wrap around to prevent integer overflows
            if csession["webclient_authenticated_nonce"] > 32:
                csession["webclient_authenticated_nonce"] = 0
